import os
import struct
import time

from urh.util.Logger import logger

from urh.signalprocessing.Message import Message


class PCAP(object):
    def __init__(self):
        self.timestamp_sec = None
        self.timestamp_nsec = None

    def reset_timestamp(self):
        self.timestamp_sec = None
        self.timestamp_nsec = None

    def build_global_header(self) -> bytes:
        MAGIC_NUMBER = 0xa1b23c4d # Nanosecond resolution
        VERSION_MAJOR, VERSION_MINOR = 2, 4
        THISZONE = 0
        SIGFIGS = 0
        SNAPLEN = 65535
        NETWORK = 147

        self.reset_timestamp()

        return struct.pack(">IHHiIII", MAGIC_NUMBER, VERSION_MAJOR, VERSION_MINOR, THISZONE, SIGFIGS, SNAPLEN, NETWORK)

    def build_packet(self, ts_sec: int, ts_nsec: int, data: bytes) -> bytes:
        if self.timestamp_nsec is None or self.timestamp_sec is None:
            self.timestamp_sec, self.timestamp_nsec = self.get_seconds_nseconds(time.time())

        self.timestamp_sec += int(ts_sec)
        self.timestamp_nsec += int(ts_nsec)
        if self.timestamp_nsec >= 1e9:
            self.timestamp_sec += int(self.timestamp_nsec / 1e9)
            self.timestamp_nsec = int(self.timestamp_nsec % 1e9)

        l = len(data)
        return struct.pack(">IIII", self.timestamp_sec, self.timestamp_nsec, l, l) + data

    def write_packets(self, packets, filename: str, sample_rate: int):
        """

        :type packets: list of Message
        :param filename:
        :return:
        """
        if os.path.isfile(filename):
            logger.warning("{0} already exists. Overwriting it".format(filename))

        with open(filename, "wb") as f:
            f.write(self.build_global_header())

        with open(filename, "ab") as f:
            rel_time_offset_ns = 0
            for pkt in packets:
                f.write(self.build_packet(0, rel_time_offset_ns, pkt.decoded_bits_buffer))
                rel_time_offset_ns = pkt.get_duration(sample_rate) * 10 ** 9

    @staticmethod
    def get_seconds_nseconds(timestamp):
        seconds = int(timestamp)
        nseconds = int((timestamp - seconds) * 10 ** 9)
        return seconds, nseconds



import xml.etree.ElementTree as ET

from urh.simulator.SimulatorItem import SimulatorItem
from urh.simulator.SimulatorRule import SimulatorRuleCondition
from urh.util.Formatter import Formatter


class SimulatorSleepAction(SimulatorItem):
    def __init__(self):
        super().__init__()
        self.sleep_time = 1.0

    @property
    def caption(self):
        return "Sleep for " + Formatter.science_time(self.sleep_time)

    def validate(self):
        return True

    def set_parent(self, value):
        if value is not None:
            assert value.parent() is None or isinstance(value, SimulatorRuleCondition)

        super().set_parent(value)

    def to_xml(self):
        attrib = {"sleep_time": str(self.sleep_time)}
        return ET.Element("simulator_sleep_action", attrib=attrib)

    @classmethod
    def from_xml(cls, tag):
        result = SimulatorSleepAction()
        result.sleep_time = Formatter.str2val(tag.get("sleep_time", "1.0"), float, 1.0)
        return result



class CommonRange(object):

    __slots__ = ["start", "end", "__bits", "__hex", "messages"]

    def __init__(self, start: int, end: int, bits: str):
        """

        :param start: Start of the common range
        :param end: End of the common range
        :param bits: Value of the common range
        """
        self.start = start
        self.end = end
        self.__bits = bits
        self.__hex = ("{0:0"+str(len(self.__bits)//4)+"x}").format(int(self.__bits, 2))
        self.messages = set()
        """:type: set of int """

    @property
    def bits(self) -> str:
        return self.__bits

    @property
    def hex_value(self) -> str:
        return self.__hex

    @property
    def byte_len(self) -> int:
        return (self.end - self.start) // 8

    def __len__(self):
        return self.end - self.start

    def __hash__(self):
        return hash(self.start) + hash(self.end) + hash(self.bits)

    def pos_of_hex(self, hex_str) -> tuple:
        try:
            start = 4 * self.hex_value.index(hex_str)
            return start, start + 4 * len(hex_str)
        except ValueError:
            return None

    @staticmethod
    def from_hex(hex_str):
        return CommonRange(start=0, end=0, bits="{0:b}".format(int(hex_str, 16)))

    def __eq__(self, other):
        if isinstance(other, CommonRange):
            return self.start == other.start and self.end == other.end and self.bits == other.bits
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, CommonRange):
            if self.start != other.start:
                return self.start < other.start
            else:
                return self.end <= other.end
        else:
            return -1

    def __repr__(self):
        return "{}-{} {} ({})".format(self.start,  self.end, self.hex_value, self.messages)
        



from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal
from PyQt5.QtGui import QKeyEvent, QCloseEvent
from PyQt5.QtWidgets import QDialog, QHeaderView, QAbstractItemView

from urh import constants
from urh.controller.widgets.ChecksumWidget import ChecksumWidget
from urh.models.PLabelTableModel import PLabelTableModel
from urh.signalprocessing.ChecksumLabel import ChecksumLabel
from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel
from urh.ui.delegates.CheckBoxDelegate import CheckBoxDelegate
from urh.ui.delegates.ComboBoxDelegate import ComboBoxDelegate
from urh.ui.delegates.SpinBoxDelegate import SpinBoxDelegate
from urh.ui.ui_properties_dialog import Ui_DialogLabels
from urh.util import util
from urh.util.Logger import logger


class ProtocolLabelDialog(QDialog):
    apply_decoding_changed = pyqtSignal(ProtocolLabel, MessageType)

    SPECIAL_CONFIG_TYPES = [FieldType.Function.CHECKSUM]

    def __init__(self, message: Message, viewtype: int, selected_index=None, parent=None):
        super().__init__(parent)
        self.ui = Ui_DialogLabels()
        self.ui.setupUi(self)
        util.set_splitter_stylesheet(self.ui.splitter)

        field_types = FieldType.load_from_xml()
        self.model = PLabelTableModel(message, field_types)

        self.ui.tblViewProtoLabels.setItemDelegateForColumn(0, ComboBoxDelegate([ft.caption for ft in field_types],
                                                                                is_editable=True,
                                                                                return_index=False, parent=self))
        self.ui.tblViewProtoLabels.setItemDelegateForColumn(1, SpinBoxDelegate(1, len(message), self))
        self.ui.tblViewProtoLabels.setItemDelegateForColumn(2, SpinBoxDelegate(1, len(message), self))
        self.ui.tblViewProtoLabels.setItemDelegateForColumn(3,
                                                            ComboBoxDelegate([""] * len(constants.LABEL_COLORS),
                                                                             colors=constants.LABEL_COLORS,
                                                                             parent=self))
        self.ui.tblViewProtoLabels.setItemDelegateForColumn(4, CheckBoxDelegate(self))
        self.ui.tblViewProtoLabels.setModel(self.model)
        self.ui.tblViewProtoLabels.setEditTriggers(QAbstractItemView.AllEditTriggers)

        self.ui.tblViewProtoLabels.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.ui.tblViewProtoLabels.resizeColumnsToContents()
        self.setWindowFlags(Qt.Window)
        self.setWindowTitle(self.tr("Edit Protocol Labels from message type %s") % message.message_type.name)

        self.configure_special_config_tabs()
        self.ui.splitter.setSizes([self.height() / 2, self.height() / 2])

        self.create_connects()

        if selected_index is not None:
            self.ui.tblViewProtoLabels.setCurrentIndex(self.model.index(selected_index, 0))

        self.ui.cbProtoView.setCurrentIndex(viewtype)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowFlags(Qt.Window)

        try:
            self.restoreGeometry(constants.SETTINGS.value("{}/geometry".format(self.__class__.__name__)))
        except TypeError:
            pass

        for i in range(self.model.rowCount()):
            self.open_editors(i)

    def configure_special_config_tabs(self):
        self.ui.tabWidgetAdvancedSettings.clear()
        for lbl in self.model.message_type:  # type: ProtocolLabel
            if isinstance(lbl, SimulatorProtocolLabel):
                lbl = lbl.label

            if lbl.field_type is not None and lbl.field_type.function in self.SPECIAL_CONFIG_TYPES:
                if isinstance(lbl, ChecksumLabel):
                    w = ChecksumWidget(lbl, self.model.message, self.model.proto_view)
                    self.ui.tabWidgetAdvancedSettings.addTab(w, lbl.name)
                else:
                    logger.error("No Special Config Dialog for field type " + lbl.field_type.caption)

        if self.ui.tabWidgetAdvancedSettings.count() > 0:
            self.ui.tabWidgetAdvancedSettings.setCurrentIndex(0)
            self.ui.tabWidgetAdvancedSettings.setFocus()

        self.ui.groupBoxAdvancedSettings.setVisible(self.ui.tabWidgetAdvancedSettings.count() > 0)

    def create_connects(self):
        self.ui.btnConfirm.clicked.connect(self.confirm)
        self.ui.cbProtoView.currentIndexChanged.connect(self.set_view_index)
        self.model.apply_decoding_changed.connect(self.on_apply_decoding_changed)
        self.model.special_status_label_changed.connect(self.on_label_special_status_changed)

    def open_editors(self, row):
        self.ui.tblViewProtoLabels.openPersistentEditor(self.model.index(row, 4))

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Enter:
            event.ignore()
        else:
            event.accept()

    def closeEvent(self, event: QCloseEvent):
        constants.SETTINGS.setValue("{}/geometry".format(self.__class__.__name__), self.saveGeometry())
        super().closeEvent(event)

    @pyqtSlot()
    def confirm(self):
        self.close()

    @pyqtSlot(int)
    def set_view_index(self, ind):
        self.model.proto_view = ind
        self.model.update()

        for i in range(self.ui.tabWidgetAdvancedSettings.count()):
            self.ui.tabWidgetAdvancedSettings.widget(i).proto_view = ind

    @pyqtSlot(ProtocolLabel)
    def on_apply_decoding_changed(self, lbl: ProtocolLabel):
        self.apply_decoding_changed.emit(lbl, self.model.message_type)

    @pyqtSlot(ProtocolLabel)
    def on_label_special_status_changed(self, lbl: ProtocolLabel):
        self.configure_special_config_tabs()



import math
from PyQt5.QtCore import Qt, QPoint, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtWidgets import QAction

from urh import constants
from urh.ui.views.EditableGraphicView import EditableGraphicView


class EpicGraphicView(EditableGraphicView):
    """
    Tied to Signal Frame (Interpretation)
    """
    save_clicked = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.save_enabled = True
        self.create_new_signal_enabled = True
        self.participants_assign_enabled = True
        self.cache_qad = True
        self.y_sep = 0

        self.save_action = QAction(self.tr("Save"), self)  # type: QAction
        self.save_action.setIcon(QIcon.fromTheme("document-save"))
        self.save_action.setShortcut(QKeySequence.Save)
        self.save_action.triggered.connect(self.on_save_action_triggered)
        self.save_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.addAction(self.save_action)

    @property
    def sample_rate(self):
        try:
            return self.signal.sample_rate
        except AttributeError:
            return None

    @sample_rate.setter
    def sample_rate(self, value):
        raise ValueError("Not implemented for epic graphic view")

    @property
    def selected_messages(self):
        if self.something_is_selected and self.protocol:
            sb, _, eb, _ = self.protocol.get_bitseq_from_selection(self.selection_area.start, abs(self.selection_area.width))
            return self.protocol.messages[sb:eb + 1]
        else:
            return []

    def is_pos_in_separea(self, pos: QPoint):
        if self.scene_type != 1:
            return False
        padding = constants.SEPARATION_PADDING * self.view_rect().height()
        return self.y_sep - padding <= pos.y() <= self.y_sep + padding

    def _get_sub_path_ranges_and_colors(self, start: float, end: float):
        sub_path_ranges = []
        colors = []
        start = max(0, int(start))
        end = int(math.ceil(end))

        if not self.protocol.messages:
            return None, None

        for message in self.protocol.messages:
            if message.bit_sample_pos[-2] < start:
                continue

            color = None if message.participant is None else constants.PARTICIPANT_COLORS[
                message.participant.color_index]

            if color is None:
                continue

            # Append the pause until first bit of message
            sub_path_ranges.append((start, message.bit_sample_pos[0]))
            if start < message.bit_sample_pos[0]:
                colors.append(None)
            else:
                colors.append(color)  # Zoomed inside a message

            if message.bit_sample_pos[-2] > end:
                sub_path_ranges.append((message.bit_sample_pos[0], end))
                colors.append(color)
                break

            # Data part of the message
            sub_path_ranges.append((message.bit_sample_pos[0], message.bit_sample_pos[-2] + 1))
            colors.append(color)

            start = message.bit_sample_pos[-2] + 1

        if sub_path_ranges and sub_path_ranges[-1][1] != end:
            sub_path_ranges.append((sub_path_ranges[-1][1], end))
            colors.append(None)

        sub_path_ranges = sub_path_ranges if sub_path_ranges else None
        colors = colors if colors else None
        return sub_path_ranges, colors

    @pyqtSlot()
    def on_save_action_triggered(self):
        self.save_clicked.emit()




import locale
import math
import os
from collections import defaultdict
from datetime import datetime

import numpy
from PyQt5.QtCore import pyqtSlot, QTimer, Qt, pyqtSignal, QItemSelection, QItemSelectionModel, QLocale, \
    QModelIndex
from PyQt5.QtGui import QContextMenuEvent, QIcon
from PyQt5.QtWidgets import QMessageBox, QAbstractItemView, QUndoStack, QMenu, QWidget, QHeaderView

from urh import constants
from urh.controller.dialogs.MessageTypeDialog import MessageTypeDialog
from urh.controller.dialogs.ProtocolLabelDialog import ProtocolLabelDialog
from urh.models.LabelValueTableModel import LabelValueTableModel
from urh.models.MessageTypeTableModel import MessageTypeTableModel
from urh.models.ParticipantListModel import ParticipantListModel
from urh.models.ProtocolTableModel import ProtocolTableModel
from urh.models.ProtocolTreeModel import ProtocolTreeModel
from urh.plugins.PluginManager import PluginManager
from urh.signalprocessing.Encoding import Encoding
from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.ProtocolGroup import ProtocolGroup
from urh.ui.delegates.ComboBoxDelegate import ComboBoxDelegate
from urh.ui.delegates.MessageTypeButtonDelegate import MessageTypeButtonDelegate
from urh.ui.ui_analysis import Ui_TabAnalysis
from urh.util import FileOperator, util
from urh.util.Formatter import Formatter
from urh.util.Logger import logger
from urh.util.ProjectManager import ProjectManager


class CompareFrameController(QWidget):
    show_interpretation_clicked = pyqtSignal(int, int, int, int)
    show_decoding_clicked = pyqtSignal()
    files_dropped = pyqtSignal(list)
    participant_changed = pyqtSignal()
    show_config_field_types_triggered = pyqtSignal()
    load_protocol_clicked = pyqtSignal()

    def __init__(self, plugin_manager: PluginManager, project_manager: ProjectManager, parent):

        super().__init__(parent)

        self.proto_analyzer = ProtocolAnalyzer(None)
        self.project_manager = project_manager

        self.ui = Ui_TabAnalysis()
        self.ui.setupUi(self)
        util.set_splitter_stylesheet(self.ui.splitter)
        util.set_splitter_stylesheet(self.ui.splitter_2)

        self.ui.lBitsSelection.setText("")
        self.ui.lDecimalSelection.setText("")
        self.ui.lHexSelection.setText("")
        self.plugin_manager = plugin_manager
        self.decimal_point = QLocale().decimalPoint()

        self.__selected_message_type = self.proto_analyzer.default_message_type

        self.participant_list_model = ParticipantListModel(project_manager.participants)
        self.ui.listViewParticipants.setModel(self.participant_list_model)

        self.__active_group_ids = [0]
        self.selected_protocols = set()

        self.search_select_filter_align_menu = QMenu()
        self.search_action = self.search_select_filter_align_menu.addAction(self.tr("Search"))
        self.select_action = self.search_select_filter_align_menu.addAction(self.tr("Select all"))
        self.filter_action = self.search_select_filter_align_menu.addAction(self.tr("Filter"))
        self.align_action = self.search_select_filter_align_menu.addAction(self.tr("Align"))
        self.ui.btnSearchSelectFilter.setMenu(self.search_select_filter_align_menu)

        self.analyze_menu = QMenu()
        self.assign_participants_action = self.analyze_menu.addAction(self.tr("Assign participants"))
        self.assign_participants_action.setCheckable(True)
        self.assign_participants_action.setChecked(True)
        self.assign_message_type_action = self.analyze_menu.addAction(self.tr("Assign message type"))
        self.assign_message_type_action.setCheckable(True)
        self.assign_message_type_action.setChecked(True)
        self.assign_labels_action = self.analyze_menu.addAction(self.tr("Assign labels"))
        self.assign_labels_action.setCheckable(True)
        self.assign_labels_action.setChecked(False)
        self.assign_participant_address_action = self.analyze_menu.addAction(self.tr("Assign participant addresses"))
        self.assign_participant_address_action.setCheckable(True)
        self.assign_participant_address_action.setChecked(True)
        self.ui.btnAnalyze.setMenu(self.analyze_menu)

        self.ui.lblShownRows.hide()
        self.ui.lblClearAlignment.hide()

        self.protocol_model = ProtocolTableModel(self.proto_analyzer, project_manager.participants,
                                                 self)  # type: ProtocolTableModel
        self.message_type_table_model = MessageTypeTableModel(self.proto_analyzer.message_types)

        self.label_value_model = LabelValueTableModel(self.proto_analyzer,
                                                      controller=self)  # type: LabelValueTableModel
        self.ui.tblViewProtocol.setModel(self.protocol_model)
        self.ui.tblViewProtocol.controller = self
        self.ui.tblLabelValues.setModel(self.label_value_model)
        self.ui.tblViewMessageTypes.setModel(self.message_type_table_model)

        self.ui.tblViewMessageTypes.setItemDelegateForColumn(1, MessageTypeButtonDelegate(
            parent=self.ui.tblViewMessageTypes))
        self.ui.tblViewMessageTypes.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.ui.tblViewMessageTypes.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeToContents)

        self.ui.tblLabelValues.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.ui.tblLabelValues.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self.ui.tblLabelValues.horizontalHeader().setSectionResizeMode(3, QHeaderView.ResizeToContents)

        self.selection_timer = QTimer(self)
        self.selection_timer.setSingleShot(True)

        self.setAcceptDrops(False)

        self.proto_tree_model = ProtocolTreeModel(controller=self)  # type: ProtocolTreeModel
        self.ui.treeViewProtocols.setModel(self.proto_tree_model)

        self.create_connects()
        self.fill_decoding_combobox()

        self.rows_for_protocols = {}
        self.__protocols = None

        self.min_height = self.minimumHeight()
        self.max_height = self.maximumHeight()

        self.old_reference_index = 0

        self.__set_decoding_error_label(None)

        self.message_type_table_model.update()

    # region properties

    @property
    def field_types(self):
        return self.project_manager.field_types

    @property
    def field_types_by_caption(self):
        return self.project_manager.field_types_by_caption

    @property
    def active_group_ids(self):
        """
        Returns a list of currently selected group indices

        :rtype: list of int
        """
        return self.__active_group_ids

    @active_group_ids.setter
    def active_group_ids(self, val: list):
        self.__active_group_ids = val

    @property
    def groups(self):
        """

        :rtype: list of ProtocolGroup
        """
        return self.proto_tree_model.groups

    @property
    def active_groups(self):
        """
        Returns a list of currently selected protocol groups

        :rtype: list of ProtocolGroup
        """
        return [self.proto_tree_model.group_at(i) for i in self.active_group_ids]

    @property
    def active_message_type(self) -> MessageType:
        return self.__selected_message_type

    @active_message_type.setter
    def active_message_type(self, val: MessageType):
        if val not in self.proto_analyzer.message_types:
            logger.error("Message type {} not in message types".format(val.name))
            return

        self.__selected_message_type = val
        self.update_field_type_combobox()

    @property
    def selected_messages(self):
        return self.ui.tblViewProtocol.selected_messages

    @property
    def protocol_undo_stack(self) -> QUndoStack:
        return self.protocol_model.undo_stack

    @property
    def protocols(self):
        """
        :rtype: dict[int, list of ProtocolAnalyzer]
        """
        if self.__protocols is None:
            self.__protocols = self.proto_tree_model.protocols
        return self.__protocols

    @property
    def protocol_list(self):
        """
        :return: visible protocols
        :rtype: list of ProtocolAnalyzer
        """
        result = []
        for group in self.groups:
            result.extend(group.protocols)
        return result

    @property
    def full_protocol_list(self):
        """
        :return: all protocols including not shown ones
        :rtype: list of ProtocolAnalyzer
        """
        result = []
        for group in self.groups:
            result.extend(group.all_protocols)
        return result

    # endregion

    def __set_decoding_error_label(self, message: Message):
        if message:
            errors = message.decoding_errors
            percent = 100 * (errors / len(message))
            state = message.decoding_state if message.decoding_state != message.decoder.ErrorState.SUCCESS else ""
            color = "green" if errors == 0 and state == "" else "red"

            self.ui.lDecodingErrorsValue.setStyleSheet("color: " + color)
            self.ui.lDecodingErrorsValue.setText(locale.format_string("%d (%.02f%%) %s", (errors, percent, state)))
        else:
            self.ui.lDecodingErrorsValue.setText("No message selected")

    def create_connects(self):
        self.protocol_undo_stack.indexChanged.connect(self.on_undo_stack_index_changed)
        self.ui.cbProtoView.currentIndexChanged.connect(self.on_protocol_view_changed)
        self.ui.cbDecoding.currentIndexChanged.connect(self.on_combobox_decoding_current_index_changed)
        self.ui.cbShowDiffs.clicked.connect(self.on_chkbox_show_differences_clicked)
        self.ui.chkBoxOnlyShowLabelsInProtocol.stateChanged.connect(self.on_check_box_show_only_labels_state_changed)
        self.ui.chkBoxShowOnlyDiffs.stateChanged.connect(self.on_check_box_show_only_diffs_state_changed)

        self.protocol_model.vertical_header_color_status_changed.connect(
            self.ui.tblViewProtocol.on_vertical_header_color_status_changed)

        self.ui.tblViewProtocol.show_interpretation_clicked.connect(self.show_interpretation_clicked.emit)
        self.ui.tblViewProtocol.protocol_view_change_clicked.connect(self.ui.cbProtoView.setCurrentIndex)
        self.ui.tblViewProtocol.selection_changed.connect(self.on_table_selection_changed)
        self.ui.tblViewProtocol.writeable_changed.connect(self.on_writeable_changed)
        self.ui.tblViewProtocol.row_visibility_changed.connect(self.on_tbl_view_protocol_row_visibility_changed)
        self.ui.tblViewProtocol.edit_label_triggered.connect(self.on_edit_label_clicked_in_table)
        self.ui.tblViewProtocol.create_label_triggered.connect(self.on_create_label_triggered)
        self.ui.tblViewProtocol.participant_changed.connect(self.on_participant_edited)
        self.ui.tblViewProtocol.messagetype_selected.connect(self.on_message_type_selected)
        self.ui.tblViewProtocol.new_messagetype_clicked.connect(self.on_table_new_message_type_clicked)
        self.ui.tblViewProtocol.files_dropped.connect(self.on_files_dropped)

        self.ui.tblLabelValues.edit_label_action_triggered.connect(self.on_edit_label_action_triggered)
        self.ui.tblLabelValues.configure_field_types_action_triggered.connect(
            self.show_config_field_types_triggered.emit)

        self.label_value_model.protolabel_visibility_changed.connect(self.on_protolabel_visibility_changed)
        self.label_value_model.protocol_label_name_edited.connect(self.label_value_model.update)
        self.label_value_model.label_removed.connect(self.on_label_removed)
        self.label_value_model.label_color_changed.connect(self.on_label_color_changed)

        self.ui.tblViewMessageTypes.selectionModel().currentRowChanged.connect(
            self.on_tbl_view_message_current_row_changed)
        self.ui.tblViewMessageTypes.configure_message_type_rules_triggered.connect(
            self.on_configure_message_type_rules_triggered)
        self.ui.tblViewMessageTypes.auto_message_type_update_triggered.connect(
            self.update_automatic_assigned_message_types)

        self.message_type_table_model.modelReset.connect(self.on_message_type_table_model_updated)
        self.message_type_table_model.message_type_removed.connect(self.on_message_type_removed)
        self.message_type_table_model.message_type_visibility_changed.connect(self.on_message_type_visibility_changed)
        self.message_type_table_model.message_type_name_edited.connect(self.on_message_type_name_edited)

        self.ui.btnSearchSelectFilter.clicked.connect(self.on_btn_search_clicked)
        self.ui.btnNextSearch.clicked.connect(self.on_btn_next_search_clicked)
        self.ui.btnPrevSearch.clicked.connect(self.on_btn_prev_search_clicked)
        self.ui.lineEditSearch.returnPressed.connect(self.ui.btnSearchSelectFilter.click)
        self.search_action.triggered.connect(self.on_search_action_triggered)
        self.select_action.triggered.connect(self.on_select_action_triggered)
        self.filter_action.triggered.connect(self.on_filter_action_triggered)
        self.align_action.triggered.connect(self.on_align_action_triggered)
        self.ui.lblShownRows.linkActivated.connect(self.on_label_shown_link_activated)
        self.ui.lblClearAlignment.linkActivated.connect(self.on_label_clear_alignment_link_activated)

        self.ui.btnSaveProto.clicked.connect(self.on_btn_save_protocol_clicked)
        self.ui.btnLoadProto.clicked.connect(self.on_btn_load_proto_clicked)

        self.ui.btnAnalyze.clicked.connect(self.on_btn_analyze_clicked)

        self.protocol_model.ref_index_changed.connect(self.on_ref_index_changed)

        self.project_manager.project_updated.connect(self.on_project_updated)
        self.participant_list_model.show_state_changed.connect(self.on_participant_show_state_changed)

        self.ui.btnAddMessagetype.clicked.connect(self.on_btn_new_message_type_clicked)

        self.selection_timer.timeout.connect(self.on_table_selection_timer_timeout)
        self.ui.treeViewProtocols.selection_changed.connect(self.on_tree_view_selection_changed)
        self.proto_tree_model.item_dropped.connect(self.on_item_in_proto_tree_dropped)

        self.proto_tree_model.group_deleted.connect(self.on_group_deleted)
        self.proto_tree_model.proto_to_group_added.connect(self.on_proto_to_group_added)

    def get_message_type_for_label(self, lbl: ProtocolLabel) -> MessageType:
        return next((msg_type for msg_type in self.proto_analyzer.message_types if lbl in msg_type), None)

    def update_field_type_combobox(self):
        field_types = [ft.caption for ft in self.field_types]
        delegate = ComboBoxDelegate(field_types, is_editable=True, return_index=False)
        self.ui.tblLabelValues.setItemDelegateForColumn(0, delegate)

    def set_decoding(self, decoding: Encoding, messages=None):
        """

        :param decoding:
        :param messages: None = set for all messages
        :return:
        """
        if decoding is None:
            self.show_decoding_clicked.emit()
        else:
            if messages is None:
                messages = self.proto_analyzer.messages
                if len(messages) > 10:
                    reply = QMessageBox.question(self, "Set decoding",
                                                 "Do you want to apply the selected decoding to {} messages?".format(
                                                     len(messages)), QMessageBox.Yes | QMessageBox.No)
                    if reply != QMessageBox.Yes:
                        self.ui.cbDecoding.blockSignals(True)
                        self.ui.cbDecoding.setCurrentText("...")
                        self.ui.cbDecoding.blockSignals(False)
                        return

            self.show_all_cols()

            for msg in messages:
                msg.decoder = decoding

            self.ui.tblViewProtocol.zero_hide_offsets.clear()
            self.clear_search()

            selected = self.ui.tblViewProtocol.selectionModel().selection()

            if not selected.isEmpty() and self.isVisible() and self.proto_analyzer.num_messages > 0:
                min_row = min(rng.top() for rng in selected)
                min_row = min_row if min_row < len(self.proto_analyzer.messages) else -1
                try:
                    msg = self.proto_analyzer.messages[min_row]
                except IndexError:
                    msg = None
                self.__set_decoding_error_label(msg)
            else:
                self.__set_decoding_error_label(None)

            self.protocol_model.update()
            self.message_type_table_model.update()
            self.label_value_model.update()

            for lbl in self.proto_analyzer.protocol_labels:
                self.set_protocol_label_visibility(lbl)

            self.ui.tblViewProtocol.resize_columns()

    @property
    def decodings(self):
        return self.project_manager.decodings

    def refresh_existing_encodings(self):
        """
        Refresh existing encodings for messages, when encoding was changed by user in dialog

        :return:
        """
        update = False

        for msg in self.proto_analyzer.messages:
            i = next((i for i, d in enumerate(self.decodings) if d.name == msg.decoder.name), 0)
            if msg.decoder != self.decodings[i]:
                update = True
                msg.decoder = self.decodings[i]
                msg.clear_decoded_bits()
                msg.clear_encoded_bits()

        if update:
            self.protocol_model.update()
            self.label_value_model.update()

    def fill_decoding_combobox(self):
        cur_item = self.ui.cbDecoding.currentText() if self.ui.cbDecoding.count() > 0 else None
        self.ui.cbDecoding.blockSignals(True)
        self.ui.cbDecoding.clear()
        prev_index = 0
        for i, decoding in enumerate(self.decodings):
            self.ui.cbDecoding.addItem(decoding.name)
            if decoding.name == cur_item:
                prev_index = i

        self.ui.cbDecoding.addItem("...")
        self.ui.cbDecoding.setCurrentIndex(prev_index)
        self.ui.cbDecoding.blockSignals(False)

    def add_protocol(self, protocol: ProtocolAnalyzer, group_id: int = 0) -> ProtocolAnalyzer:
        self.__protocols = None
        self.proto_tree_model.add_protocol(protocol, group_id)
        protocol.qt_signals.protocol_updated.connect(self.on_protocol_updated)
        if protocol.signal:
            protocol.signal.sample_rate_changed.connect(self.set_shown_protocols)  # Refresh times
        protocol.qt_signals.show_state_changed.connect(self.set_shown_protocols)
        protocol.qt_signals.show_state_changed.connect(self.filter_search_results)
        for i in range(self.proto_tree_model.ngroups):
            self.expand_group_node(i)
        return protocol

    def add_protocol_from_file(self, filename: str) -> ProtocolAnalyzer:
        pa = ProtocolAnalyzer(signal=None, filename=filename)
        pa.message_types = []

        if filename.endswith(".bin"):
            pa.from_binary(filename)
        else:
            pa.from_xml_file(filename=filename, read_bits=True)
            for messsage_type in pa.message_types:
                if messsage_type not in self.proto_analyzer.message_types:
                    if messsage_type.name in (mt.name for mt in self.proto_analyzer.message_types):
                        messsage_type.name += " (" + os.path.split(filename)[1].rstrip(".xml").rstrip(".proto") + ")"
                    self.proto_analyzer.message_types.append(messsage_type)

        self.message_type_table_model.update()
        self.add_protocol(protocol=pa)

        self.set_shown_protocols()
        return pa

    def add_sniffed_protocol_messages(self, messages: list):
        if len(messages) > 0:
            proto_analyzer = ProtocolAnalyzer(None)
            proto_analyzer.name = datetime.fromtimestamp(messages[0].timestamp).strftime("%Y-%m-%d %H:%M:%S")
            proto_analyzer.messages = messages
            self.add_protocol(proto_analyzer, group_id=self.proto_tree_model.ngroups - 1)
            self.refresh()

    def add_protocol_label(self, start: int, end: int, messagenr: int, proto_view: int, edit_label_name=True):
        # Ensure atleast one Group is active
        start, end = self.proto_analyzer.convert_range(start, end, proto_view, 0, decoded=True, message_indx=messagenr)
        message_type = self.proto_analyzer.messages[messagenr].message_type
        try:
            used_field_types = [lbl.field_type for lbl in message_type]
            first_unused_type = next(ft for ft in self.field_types if ft not in used_field_types)
            name = first_unused_type.caption
        except (StopIteration, AttributeError):
            first_unused_type, name = None, None

        proto_label = message_type.add_protocol_label(start=start, end=end, name=name, type=first_unused_type)

        self.message_type_table_model.update()
        self.protocol_model.update()
        self.label_value_model.update()

        if edit_label_name:
            try:
                index = self.ui.tblLabelValues.model().index(message_type.index(proto_label), 0)
                self.ui.tblLabelValues.setCurrentIndex(index)
                self.ui.tblLabelValues.edit(index)
            except ValueError:
                pass

        return True

    def add_message_type(self, selected_messages: list = None):
        selected_messages = selected_messages if isinstance(selected_messages, list) else []
        self.proto_analyzer.add_new_message_type(labels=self.proto_analyzer.default_message_type)
        self.message_type_table_model.update()
        self.active_message_type = self.proto_analyzer.message_types[-1]
        for msg in selected_messages:
            msg.message_type = self.active_message_type
        self.protocol_model.update()

    def remove_protocol(self, protocol: ProtocolAnalyzer):
        self.__protocols = None
        self.proto_tree_model.remove_protocol(protocol)
        try:
            del self.rows_for_protocols[protocol]
        except KeyError:
            pass
        self.ui.tblViewProtocol.clearSelection()
        self.set_shown_protocols()

    def set_shown_protocols(self):
        hidden_rows = {i for i in range(self.protocol_model.row_count) if self.ui.tblViewProtocol.isRowHidden(i)}
        relative_hidden_row_positions = {}
        for proto in self.rows_for_protocols.keys():
            if any(i in hidden_rows for i in self.rows_for_protocols[proto]):
                m = min(self.rows_for_protocols[proto])
                relative_hidden_row_positions[proto] = [i - m for i in hidden_rows if
                                                        i in self.rows_for_protocols[proto]]

        # self.protocol_undo_stack.clear()
        self.proto_analyzer.messages[:] = []
        self.rows_for_protocols.clear()
        align_labels = constants.SETTINGS.value("align_labels", True, bool)
        line = 0
        first_msg_indices = []
        prev_line = 0
        for proto in self.protocol_list:
            abs_time = 0
            rel_time = 0
            if proto.show and proto.messages:
                num_messages = 0
                for i, message in enumerate(proto.messages):
                    if not message:
                        continue

                    message.align_labels = align_labels
                    try:
                        if hasattr(proto.signal, "sample_rate"):
                            if i > 0:
                                rel_time = proto.messages[i - 1].get_duration(proto.signal.sample_rate)
                                abs_time += rel_time
                        else:
                            # No signal, loaded from protocol file
                            abs_time = datetime.fromtimestamp(message.timestamp).strftime("%Y-%m-%d %H:%M:%S.%f")
                            if i > 0:
                                rel_time = message.timestamp - proto.messages[i - 1].timestamp
                    except IndexError:
                        pass

                    message.absolute_time = abs_time
                    message.relative_time = rel_time

                    num_messages += 1
                    if message.message_type not in self.proto_analyzer.message_types:
                        message.message_type = self.proto_analyzer.default_message_type
                    self.proto_analyzer.messages.append(message)

                line += num_messages
                rows_for_cur_proto = list(range(prev_line, line))
                self.rows_for_protocols[proto] = rows_for_cur_proto[:]

                prev_line = line
                if line != 0:
                    first_msg_indices.append(line)

        # apply hidden rows to new order
        for i in range(self.protocol_model.row_count):
            self.ui.tblViewProtocol.showRow(i)

        self.protocol_model.hidden_rows.clear()
        for proto in relative_hidden_row_positions.keys():
            try:
                start = min(self.rows_for_protocols[proto])
                for rel_pos in relative_hidden_row_positions[proto]:
                    self.ui.tblViewProtocol.hideRow(start + rel_pos)
                    self.protocol_model.hidden_rows.add(start + rel_pos)
            except (KeyError, ValueError):
                pass

        # consider hidden rows
        for i in range(self.protocol_model.row_count):
            if self.ui.tblViewProtocol.isRowHidden(i) and i in first_msg_indices:
                indx = first_msg_indices.index(i)
                first_msg_indices[indx] += 1
                try:
                    if first_msg_indices[indx] >= first_msg_indices[indx + 1]:
                        del first_msg_indices[indx]
                except IndexError:
                    pass

        for line in first_msg_indices:
            self.ui.tblViewProtocol.setRowHeight(line, constants.SEPARATION_ROW_HEIGHT)

        self.protocol_model.first_messages = first_msg_indices[:]

        self.updateUI()
        self.show_differences(self.ui.cbShowDiffs.isChecked())

    def restore_selection(self, old_view: int, sel_cols, sel_rows):
        if len(sel_cols) == 0 or len(sel_rows) == 0:
            return

        start_col, end_col = numpy.min(sel_cols), numpy.max(sel_cols)
        start_row, end_row = numpy.min(sel_rows), numpy.max(sel_rows)
        new_view = self.ui.cbProtoView.currentIndex()

        message = self.proto_analyzer.messages[end_row]
        start_col = message.convert_index(start_col, old_view, new_view, True)[0]
        end_col = message.convert_index(end_col, old_view, new_view, True)[1]

        start_index = self.protocol_model.index(start_row, start_col)
        end_index = self.protocol_model.index(end_row, end_col)
        mid_index = self.protocol_model.index(int((start_row + end_row) / 2), int((start_col + end_col) / 2))

        sel = QItemSelection()
        sel.select(start_index, end_index)

        self.ui.tblViewProtocol.selectionModel().select(sel, QItemSelectionModel.ClearAndSelect)
        self.ui.tblViewProtocol.scrollTo(mid_index)

    def expand_group_node(self, group_id):
        index = self.proto_tree_model.createIndex(group_id, 0, self.proto_tree_model.rootItem.child(group_id))
        self.ui.treeViewProtocols.expand(index)

    def updateUI(self, ignore_table_model=False, resize_table=True):
        if not ignore_table_model:
            self.protocol_model.update()

        self.proto_tree_model.layoutChanged.emit()  # do not call update, as it prevents editing
        self.label_value_model.update()
        self.message_type_table_model.update()

        if resize_table:
            self.ui.tblViewProtocol.resize_columns()

    def refresh(self):
        self.__protocols = None
        self.set_shown_protocols()

    def reset(self):
        self.proto_tree_model.rootItem.clearChilds()
        self.proto_tree_model.rootItem.addGroup()
        self.refresh()

    def create_protocol_label_dialog(self, selected_index=None):
        view_type = self.ui.cbProtoView.currentIndex()
        try:
            longest_message = max(
                (msg for msg in self.proto_analyzer.messages if msg.message_type == self.active_message_type), key=len)
        except ValueError:
            logger.warning("Configuring message type with empty message set.")
            longest_message = Message([True] * 1000, 1000, self.active_message_type)
        protocol_label_dialog = ProtocolLabelDialog(message=longest_message, viewtype=view_type,
                                                    selected_index=selected_index, parent=self)
        protocol_label_dialog.apply_decoding_changed.connect(self.on_apply_decoding_changed)
        protocol_label_dialog.finished.connect(self.on_protocol_label_dialog_finished)

        return protocol_label_dialog

    def show_protocol_label_dialog(self, selected_index=None):
        dialog = self.create_protocol_label_dialog(selected_index=selected_index)
        dialog.exec_()

    def search(self):
        value = self.ui.lineEditSearch.text()
        nresults = self.protocol_model.find_protocol_value(value)

        if nresults > 0:
            self.ui.btnNextSearch.setEnabled(True)
            self.ui.btnPrevSearch.setEnabled(False)
            self.ui.lSearchTotal.setText(str(nresults))
            self.ui.lSearchCurrent.setText("0")
            self.next_search_result()
        else:
            self.clear_search()

    def select_all_search_results(self):
        self.search()
        self.ui.tblViewProtocol.clearSelection()

        for search_result in self.protocol_model.search_results:
            startindex = self.protocol_model.index(search_result[0], search_result[1])
            endindex = self.protocol_model.index(search_result[0],
                                                 search_result[1] + len(self.protocol_model.search_value) - 1)

            sel = QItemSelection()
            sel.select(startindex, endindex)

            self.ui.tblViewProtocol.selectionModel().select(sel, QItemSelectionModel.Select)
            self.ui.tblViewProtocol.scrollTo(startindex, QAbstractItemView.PositionAtCenter)

        self.ui.tblViewProtocol.setFocus()

    def filter_search_results(self):
        if "Filter" not in self.ui.btnSearchSelectFilter.text():
            # Checking for equality is not enough as some desktop environments (I am watching at you KDE!)
            # insert a & at beginning of the string
            return

        self.setCursor(Qt.WaitCursor)
        if self.ui.lineEditSearch.text():
            self.search()
            self.ui.tblLabelValues.clearSelection()

            matching_rows = set(search_result[0] for search_result in self.protocol_model.search_results)
            rows_to_hide = set(range(0, self.protocol_model.row_count)) - matching_rows
            self.ui.tblViewProtocol.hide_rows(rows_to_hide)
        else:
            self.show_all_rows()
            self.set_shown_protocols()

        self.unsetCursor()

    def __set_shown_rows_status_label(self):
        if len(self.protocol_model.hidden_rows) > 0:
            rc = self.protocol_model.row_count
            text = self.tr("shown: {}/{} (<a href='reset_filter'>reset</a>)")
            self.ui.lblShownRows.setText(text.format(rc - len(self.protocol_model.hidden_rows), rc))
            self.ui.lblShownRows.show()
        else:
            self.ui.lblShownRows.hide()

    def align_messages(self, pattern=None):
        pattern = self.ui.lineEditSearch.text() if pattern is None else pattern
        self.proto_analyzer.align_messages(pattern, view_type=self.ui.cbProtoView.currentIndex())
        self.ui.lblClearAlignment.setVisible(any(msg.alignment_offset != 0 for msg in self.proto_analyzer.messages))
        self.protocol_model.update()

        row = column = 0
        for i, message in enumerate(self.proto_analyzer.messages):
            if self.ui.tblViewProtocol.isRowHidden(i):
                continue

            data = message.view_to_string(self.ui.cbProtoView.currentIndex(), decoded=True)
            try:
                row = i
                column = data.index(pattern) + len(pattern) + self.protocol_model.get_alignment_offset_at(i) - 1
                break
            except ValueError:
                pass

        self.ui.tblViewProtocol.scrollTo(self.protocol_model.index(row, column))

        self.show_all_cols()
        for lbl in filter(lambda l: not l.show, self.proto_analyzer.protocol_labels):
            self.set_protocol_label_visibility(lbl)

    def next_search_result(self):
        index = int(self.ui.lSearchCurrent.text())
        self.ui.lSearchTotal.setText((str(len(self.protocol_model.search_results))))
        try:
            search_result = self.protocol_model.search_results[index]
            startindex = self.protocol_model.index(search_result[0], search_result[1])
            endindex = self.protocol_model.index(search_result[0],
                                                 search_result[1] + len(self.protocol_model.search_value) - 1)

            sel = QItemSelection()
            sel.select(startindex, endindex)

            self.ui.tblViewProtocol.selectionModel().select(sel, QItemSelectionModel.ClearAndSelect)
            self.ui.tblViewProtocol.scrollTo(startindex, QAbstractItemView.PositionAtCenter)

            self.ui.lSearchCurrent.setText(str(index + 1))
        except IndexError:
            self.ui.lSearchCurrent.setText("-")
        finally:
            if index + 1 == len(self.protocol_model.search_results):
                self.ui.btnNextSearch.setEnabled(False)

            if index > 0 and len(self.protocol_model.search_results) > 0:
                self.ui.btnPrevSearch.setEnabled(True)

    def prev_search_result(self):
        index = int(self.ui.lSearchCurrent.text()) - 2

        try:
            search_result = self.protocol_model.search_results[index]
            startindex = self.protocol_model.index(search_result[0], search_result[1])
            endindex = self.protocol_model.index(search_result[0],
                                                 search_result[1] + len(self.protocol_model.search_value) - 1)

            sel = QItemSelection()
            sel.select(startindex, endindex)

            self.ui.tblViewProtocol.selectionModel().select(sel, QItemSelectionModel.ClearAndSelect)
            self.ui.tblViewProtocol.scrollTo(startindex, QAbstractItemView.PositionAtCenter)

            self.ui.lSearchCurrent.setText(str(index + 1))
        except IndexError:
            self.ui.lSearchCurrent.setText("-")
        finally:
            if self.ui.lSearchCurrent.text() == "1":
                self.ui.btnPrevSearch.setEnabled(False)

            if len(self.protocol_model.search_results) > 0:
                self.ui.btnNextSearch.setEnabled(True)

    def clear_search(self):
        self.ui.btnPrevSearch.setEnabled(False)
        self.ui.btnNextSearch.setEnabled(False)
        self.ui.lSearchTotal.setText("-")
        self.ui.lSearchCurrent.setText("-")
        self.protocol_model.search_results[:] = []
        self.protocol_model.search_value = ""

    def set_protocol_label_visibility(self, lbl: ProtocolLabel, message: Message = None):
        try:
            message = message if message else next(m for m in self.proto_analyzer.messages if lbl in m.message_type)
            start, end = message.get_label_range(lbl, self.ui.cbProtoView.currentIndex(), True, consider_alignment=True)

            for i in range(start, end):
                self.ui.tblViewProtocol.setColumnHidden(i, not lbl.show)
        except Exception as e:
            pass

    def set_message_type_visibility(self, message_type: MessageType):
        try:
            rows = {i for i, msg in enumerate(self.proto_analyzer.messages) if msg.message_type == message_type}
            if message_type.show:
                self.ui.tblViewProtocol.show_rows(rows)
            else:
                self.ui.tblViewProtocol.hide_rows(rows)
        except Exception as e:
            logger.exception(e)

    def show_all_rows(self):
        self.ui.lblShownRows.hide()
        self.ui.tblViewProtocol.show_rows(range(0, self.protocol_model.row_count))
        self.set_shown_protocols()

    def show_all_cols(self):
        for i in range(self.protocol_model.col_count):
            self.ui.tblViewProtocol.showColumn(i)

    def save_protocol(self):
        for msg in self.proto_analyzer.messages:
            if not msg.decoder.is_nrz:
                reply = QMessageBox.question(self, "Saving of protocol",
                                             "You want to save this protocol with an encoding different from NRZ.\n"
                                             "This may cause loss of information if you load it again.\n\n"
                                             "Save anyway?", QMessageBox.Yes | QMessageBox.No)
                if reply != QMessageBox.Yes:
                    return
                else:
                    break

        text = "protocol"
        filename = FileOperator.get_save_file_name("{0}.proto.xml".format(text), caption="Save protocol")

        if not filename:
            return

        if filename.endswith(".bin"):
            self.proto_analyzer.to_binary(filename, use_decoded=True)
        else:
            self.proto_analyzer.to_xml_file(filename=filename, decoders=self.decodings,
                                            participants=self.project_manager.participants, write_bits=True)

    def show_differences(self, show_differences: bool):
        if show_differences:
            if self.protocol_model.refindex == -1:
                self.protocol_model.refindex = self.old_reference_index
        else:
            if self.protocol_model.refindex != -1:
                self.old_reference_index = self.protocol_model.refindex

            self.ui.chkBoxShowOnlyDiffs.setChecked(False)
            self.protocol_model.refindex = -1

    def set_show_only_status(self):
        """
        Handles the different combinations of the show only checkboxes, namely:
         - Show only labels
         - Show only Diffs
        """
        if self.ui.chkBoxShowOnlyDiffs.isChecked() and not self.ui.cbShowDiffs.isChecked():
            self.ui.cbShowDiffs.setChecked(True)
            self.show_differences(True)

        if self.ui.chkBoxOnlyShowLabelsInProtocol.isChecked() and self.ui.chkBoxShowOnlyDiffs.isChecked():
            self.show_only_diffs_and_labels()
        elif self.ui.chkBoxOnlyShowLabelsInProtocol.isChecked() and not self.ui.chkBoxShowOnlyDiffs.isChecked():
            self.show_only_labels()
        elif not self.ui.chkBoxOnlyShowLabelsInProtocol.isChecked() and self.ui.chkBoxShowOnlyDiffs.isChecked():
            self.show_only_diffs()
        else:
            self.restore_visibility()

        self.ui.tblViewProtocol.resize_columns()

    def show_only_labels(self):
        visible_columns = set()
        for msg in self.proto_analyzer.messages:
            for lbl in filter(lambda lbl: lbl.show, msg.message_type):
                start, end = msg.get_label_range(lbl=lbl, view=self.ui.cbProtoView.currentIndex(), decode=True)
                visible_columns |= set(range(start, end))

        for i in range(self.protocol_model.col_count):
            self.ui.tblViewProtocol.setColumnHidden(i, i not in visible_columns)

    def show_only_diffs(self):
        visible_rows = [i for i in range(self.protocol_model.row_count) if not self.ui.tblViewProtocol.isRowHidden(i)
                        and i != self.protocol_model.refindex]

        visible_diff_columns = [diff_col for i in visible_rows for diff_col in self.protocol_model.diff_columns[i]]

        for j in range(self.protocol_model.col_count):
            if j in visible_diff_columns:
                self.ui.tblViewProtocol.showColumn(j)
            else:
                self.ui.tblViewProtocol.hideColumn(j)

    def show_only_diffs_and_labels(self):
        visible_label_columns = set()
        for lbl in self.proto_analyzer.protocol_labels:
            if lbl.show:
                start, end = self.proto_analyzer.messages[0].get_label_range(lbl, self.ui.cbProtoView.currentIndex(),
                                                                             True)
                visible_label_columns |= (set(range(start, end)))

        visible_rows = [i for i in range(self.protocol_model.row_count) if not self.ui.tblViewProtocol.isRowHidden(i)
                        and i != self.protocol_model.refindex]

        visible_diff_columns = set([diff_col for i in visible_rows for diff_col in self.protocol_model.diff_columns[i]])

        visible_cols = visible_label_columns & visible_diff_columns
        for j in range(self.protocol_model.col_count):
            if j in visible_cols:
                self.ui.tblViewProtocol.showColumn(j)
            else:
                self.ui.tblViewProtocol.hideColumn(j)

    def restore_visibility(self):
        selected = self.ui.tblViewProtocol.selectionModel().selection()  # type: QItemSelection

        for i in range(self.protocol_model.col_count):
            self.ui.tblViewProtocol.showColumn(i)

        for lbl in filter(lambda lbl: not lbl.show, self.proto_analyzer.protocol_labels):
            self.set_protocol_label_visibility(lbl)

        if not selected.isEmpty():
            min_row = numpy.min([rng.top() for rng in selected])
            start = numpy.min([rng.left() for rng in selected])
            self.ui.tblViewProtocol.scrollTo(self.protocol_model.index(min_row, start))

    def get_labels_from_selection(self, row_start: int, row_end: int, col_start: int, col_end: int):
        """

        :rtype: list of ProtocolLabel
        """
        row_end += 1
        col_end += 1

        view = self.ui.cbProtoView.currentIndex()
        result = []
        f = 1 if self.ui.cbProtoView.currentIndex() == 0 else 4 if self.ui.cbProtoView.currentIndex() == 1 else 8
        for i in range(row_start, row_end):
            message = self.proto_analyzer.messages[i]
            for label in message.message_type:
                if label in result:
                    continue
                lbl_start, lbl_end = message.get_label_range(lbl=label, view=view, decode=True)
                a = message.alignment_offset
                if any(j in range(lbl_start, lbl_end) for j in range(col_start - a//f, col_end - a//f)):
                    result.append(label)

        return result

    def set_search_ui_visibility(self, visible: bool):
        self.ui.btnPrevSearch.setVisible(visible)
        self.ui.lSearchCurrent.setVisible(visible)
        self.ui.lSlash.setVisible(visible)
        self.ui.lSearchTotal.setVisible(visible)
        self.ui.btnNextSearch.setVisible(visible)

    def update_automatic_assigned_message_types(self):
        self.proto_analyzer.update_auto_message_types()
        self.message_type_table_model.update()
        self.protocol_model.update()

    def refresh_assigned_participants_ui(self):
        self.protocol_model.refresh_vertical_header()
        self.ui.tblViewProtocol.resize_vertical_header()
        self.participant_changed.emit()

    def refresh_field_types_for_labels(self):
        for mt in self.proto_analyzer.message_types:
            for lbl in (lbl for lbl in mt if lbl.field_type is not None):  # type: ProtocolLabel
                mt.change_field_type_of_label(lbl, self.field_types_by_caption.get(lbl.field_type.caption, None))

        self.update_field_type_combobox()

    def mousePressEvent(self, event):
        return

    def contextMenuEvent(self, event: QContextMenuEvent):
        pass

    @pyqtSlot(int)
    def on_protocol_label_dialog_finished(self, dialog_result: int):
        self.message_type_table_model.update()
        self.update_field_type_combobox()
        self.label_value_model.update()
        self.show_all_cols()
        for lbl in self.proto_analyzer.protocol_labels:
            self.set_protocol_label_visibility(lbl)
        self.set_show_only_status()
        self.protocol_model.update()
        self.ui.tblViewProtocol.resize_columns()

    @pyqtSlot()
    def on_btn_analyze_clicked(self):
        self.setCursor(Qt.WaitCursor)
        self.ui.stackedWidgetLogicAnalysis.setCurrentIndex(1)

        self.ui.progressBarLogicAnalyzer.setFormat("%p% (Detecting participants)")
        self.ui.progressBarLogicAnalyzer.setValue(0)

        if self.assign_participants_action.isChecked():
            for protocol in self.protocol_list:
                protocol.auto_assign_participants(self.protocol_model.participants)
            self.refresh_assigned_participants_ui()

        self.ui.progressBarLogicAnalyzer.setFormat("%p% (Assign message type by rules)")
        self.ui.progressBarLogicAnalyzer.setValue(50)

        if self.assign_message_type_action.isChecked():
            self.update_automatic_assigned_message_types()

        self.ui.progressBarLogicAnalyzer.setFormat("%p% (Find new labels/message types)")
        self.ui.progressBarLogicAnalyzer.setValue(75)

        if self.assign_labels_action.isChecked():
            self.proto_analyzer.auto_assign_labels()
            self.protocol_model.update()
            self.label_value_model.update()
            self.message_type_table_model.update()
            self.ui.tblViewMessageTypes.clearSelection()

        self.ui.progressBarLogicAnalyzer.setValue(90)

        if self.assign_participant_address_action.isChecked():
            self.proto_analyzer.auto_assign_participant_addresses(self.protocol_model.participants)

        self.ui.progressBarLogicAnalyzer.setValue(100)
        self.unsetCursor()
        self.ui.stackedWidgetLogicAnalysis.setCurrentIndex(0)

        self.message_type_table_model.update()  # in case message types were added by logic analyzer

    @pyqtSlot()
    def on_btn_save_protocol_clicked(self):
        self.save_protocol()

    @pyqtSlot()
    def on_btn_load_proto_clicked(self):
        self.load_protocol_clicked.emit()

    @pyqtSlot()
    def on_btn_next_search_clicked(self):
        self.next_search_result()

    @pyqtSlot()
    def on_btn_prev_search_clicked(self):
        self.prev_search_result()

    @pyqtSlot()
    def on_btn_search_clicked(self):
        self.search()

    @pyqtSlot(int)
    def on_configure_message_type_rules_triggered(self, message_type_index: int):
        dialog = MessageTypeDialog(self.proto_analyzer.message_types[message_type_index], parent=self)
        dialog.show()
        dialog.accepted.connect(self.on_message_type_dialog_accepted)

    @pyqtSlot()
    def on_btn_new_message_type_clicked(self):
        self.add_message_type()

    @pyqtSlot(list)
    def on_table_new_message_type_clicked(self, selected_messages: list):
        self.add_message_type(selected_messages)

    @pyqtSlot()
    def on_table_selection_changed(self):
        self.selection_timer.start(0)

    @pyqtSlot()
    def on_combobox_decoding_current_index_changed(self):
        new_index = self.ui.cbDecoding.currentIndex()
        if new_index == -1:
            return

        if new_index == self.ui.cbDecoding.count() - 1:
            self.set_decoding(None)
        else:
            self.set_decoding(self.decodings[new_index],
                              messages=self.selected_messages if self.selected_messages else None)

    @pyqtSlot()
    def on_participant_show_state_changed(self):
        for i, msg in enumerate(self.proto_analyzer.messages):
            hide = not msg.participant.show if msg.participant is not None else not self.participant_list_model.show_unassigned
            self.ui.tblViewProtocol.setRowHidden(i, hide)

        self.set_shown_protocols()

    @pyqtSlot(int, int)
    def on_group_deleted(self, deleted_group_id: int, new_group_id_of_childs: int):
        try:
            self.active_group_ids.remove(deleted_group_id)
        except:
            pass

        self.updateUI()

    @pyqtSlot()
    def on_tbl_view_protocol_row_visibility_changed(self):
        self.__set_shown_rows_status_label()
        self.set_shown_protocols()
        self.set_show_only_status()

        states_by_message_type = defaultdict(list)
        for i, msg in enumerate(self.proto_analyzer.messages):
            state = Qt.Unchecked if i in self.protocol_model.hidden_rows else Qt.Checked
            states_by_message_type[msg.message_type].append(state)

        for msg_type, states in states_by_message_type.items():
            msg_type.show = states[0] if len(set(states)) == 1 else Qt.PartiallyChecked

        self.message_type_table_model.update()

    @pyqtSlot(int)
    def on_check_box_show_only_labels_state_changed(self, new_state: int):
        self.set_show_only_status()

    @pyqtSlot(int)
    def on_check_box_show_only_diffs_state_changed(self, new_state: int):
        self.set_show_only_status()

    @pyqtSlot(int)
    def on_proto_to_group_added(self, group_id: int):
        self.expand_group_node(group_id)

    @pyqtSlot(ProtocolLabel)
    def on_protolabel_visibility_changed(self, proto_label: ProtocolLabel):
        self.set_protocol_label_visibility(proto_label)
        self.label_value_model.update()

    @pyqtSlot(list)
    def on_files_dropped(self, files: list):
        self.files_dropped.emit(files)

    @pyqtSlot(bool)
    def on_chkbox_show_differences_clicked(self, checked: bool):
        self.show_differences(checked)

    @pyqtSlot()
    def on_search_action_triggered(self):
        self.ui.btnSearchSelectFilter.setText("Search")
        self.ui.btnSearchSelectFilter.setIcon(QIcon.fromTheme("edit-find"))
        self.set_search_ui_visibility(True)
        self.ui.btnSearchSelectFilter.clicked.disconnect()
        self.ui.btnSearchSelectFilter.clicked.connect(self.on_btn_search_clicked)

    @pyqtSlot()
    def on_select_action_triggered(self):
        self.ui.btnSearchSelectFilter.setText("Select all")
        self.ui.btnSearchSelectFilter.setIcon(QIcon.fromTheme("edit-select-all"))
        self.set_search_ui_visibility(False)
        self.ui.btnSearchSelectFilter.clicked.disconnect()
        self.ui.btnSearchSelectFilter.clicked.connect(self.select_all_search_results)

    @pyqtSlot()
    def on_filter_action_triggered(self):
        self.ui.btnSearchSelectFilter.setText("Filter")
        self.ui.btnSearchSelectFilter.setIcon(QIcon.fromTheme("view-filter"))
        self.set_search_ui_visibility(False)
        self.ui.btnSearchSelectFilter.clicked.disconnect()
        self.ui.btnSearchSelectFilter.clicked.connect(self.filter_search_results)

    @pyqtSlot()
    def on_align_action_triggered(self):
        def on_btn_search_select_filter_clicked():
            self.align_messages()

        self.ui.btnSearchSelectFilter.setText("Align")
        self.ui.btnSearchSelectFilter.setIcon(QIcon.fromTheme("align-horizontal-left"))
        self.set_search_ui_visibility(False)
        self.ui.btnSearchSelectFilter.clicked.disconnect()
        self.ui.btnSearchSelectFilter.clicked.connect(on_btn_search_select_filter_clicked)

    @pyqtSlot(bool)
    def on_writeable_changed(self, writeable_status: bool):
        hidden_rows = {i for i in range(self.protocol_model.row_count) if self.ui.tblViewProtocol.isRowHidden(i)}
        self.protocol_model.is_writeable = writeable_status
        self.proto_tree_model.set_copy_mode(writeable_status)
        self.ui.cbDecoding.setDisabled(writeable_status)
        self.refresh()
        self.ui.tblViewProtocol.hide_rows(hidden_rows)

    @pyqtSlot()
    def on_project_updated(self):
        self.participant_list_model.update()
        self.protocol_model.refresh_vertical_header()
        self.active_message_type = self.proto_analyzer.default_message_type

    @pyqtSlot(ProtocolLabel)
    def on_label_removed(self, plabel: ProtocolLabel):
        if not plabel.show:
            self.show_all_cols()
            for lbl in self.proto_analyzer.protocol_labels:
                self.set_protocol_label_visibility(lbl)

            self.ui.tblViewProtocol.resize_columns()

        self.protocol_model.update()
        self.label_value_model.update()

    @pyqtSlot(ProtocolLabel, MessageType)
    def on_apply_decoding_changed(self, lbl: ProtocolLabel, message_type: MessageType):
        for msg in self.proto_analyzer.messages:
            if msg.message_type == message_type:
                msg.clear_decoded_bits()
                msg.clear_encoded_bits()

    @pyqtSlot(int)
    def on_undo_stack_index_changed(self, index: int):
        self.protocol_model.update()
        self.message_type_table_model.update()
        self.search()

    @pyqtSlot(int)
    def on_edit_label_clicked_in_table(self, proto_label_index: int):
        try:
            self.show_protocol_label_dialog(proto_label_index)
        except AttributeError:
            self.show_protocol_label_dialog(None)

    @pyqtSlot(int, int, int)
    def on_create_label_triggered(self, msg_index, start, end):
        a = self.protocol_model.get_alignment_offset_at(msg_index)
        self.add_protocol_label(start=start - a, end=end - a - 1,
                                messagenr=msg_index, proto_view=self.ui.cbProtoView.currentIndex())

    @pyqtSlot()
    def on_edit_label_action_triggered(self):
        self.show_protocol_label_dialog()

    @pyqtSlot()
    def on_protocol_view_changed(self):
        old_view = self.protocol_model.proto_view
        selected_indexes = self.ui.tblViewProtocol.selectionModel().selectedIndexes()
        sel_cols = [index.column() for index in selected_indexes]
        sel_rows = [index.row() for index in selected_indexes]

        self.show_all_cols()
        self.protocol_model.proto_view = self.ui.cbProtoView.currentIndex()
        self.clear_search()

        for lbl in self.proto_analyzer.protocol_labels:
            self.set_protocol_label_visibility(lbl)

        self.set_show_only_status()

        self.restore_selection(old_view, sel_cols, sel_rows)

    @pyqtSlot()
    def on_item_in_proto_tree_dropped(self):
        self.__protocols = None
        self.set_shown_protocols()
        self.ui.treeViewProtocols.clearSelection()

    @pyqtSlot()
    def on_tree_view_selection_changed(self):
        indexes = self.ui.treeViewProtocols.selectedIndexes()

        selected_items = [self.proto_tree_model.getItem(index) for index in indexes]

        self.ui.tblViewProtocol.blockSignals(True)
        active_group_ids = set()
        sel = QItemSelection()

        for item in selected_items:
            if item.is_group:
                active_group_ids.add(self.proto_tree_model.rootItem.index_of(item))
            elif item.show:
                active_group_ids.add(self.proto_tree_model.rootItem.index_of(item.parent()))

        if len(active_group_ids) == 0:
            active_group_ids.add(0)

        if active_group_ids == set(self.active_group_ids):
            ignore_table_model_on_update = True
        else:
            ignore_table_model_on_update = False
            self.active_group_ids = list(active_group_ids)
            self.active_group_ids.sort()

        self.ui.tblViewProtocol.selectionModel().select(sel, QItemSelectionModel.ClearAndSelect)
        self.ui.tblViewProtocol.blockSignals(False)

        self.updateUI(ignore_table_model=ignore_table_model_on_update)

    @pyqtSlot()
    def on_table_selection_timer_timeout(self):
        self.label_value_model.show_label_values = True
        min_row, max_row, start, end = self.ui.tblViewProtocol.selection_range()
        self.selected_protocols.clear()
        self.active_group_ids.clear()
        self.message_type_table_model.selected_message_type_indices.clear()
        self.label_value_model.selected_label_indices.clear()

        if min_row == max_row == start == end == -1:
            self.label_value_model.message_index = -1
            self.active_message_type = self.proto_analyzer.default_message_type
            self.ui.lBitsSelection.setText("")
            self.ui.lDecimalSelection.setText("")
            self.ui.lHexSelection.setText("")
            self.ui.lNumSelectedColumns.setText("0")
            self.ui.lblLabelValues.setText(self.tr("Labels of {}".format(self.active_message_type.name)))
            self.__set_decoding_error_label(message=None)
            self.updateUI(ignore_table_model=True, resize_table=False)
            return -1, -1

        selected_messages = self.selected_messages
        self.message_type_table_model.selected_message_type_indices = {
            self.proto_analyzer.message_types.index(msg.message_type) for msg in selected_messages
        }

        cur_view = self.ui.cbProtoView.currentIndex()
        self.ui.lNumSelectedColumns.setText(str(end - start))

        message = self.proto_analyzer.messages[min_row]
        self.active_message_type = message.message_type

        selected_labels = self.get_labels_from_selection(min_row, min_row, start, end - 1)
        self.label_value_model.selected_label_indices = {
            self.active_message_type.index(lbl) for lbl in selected_labels
        }

        f = 4 if cur_view == 1 else 8 if cur_view == 2 else 1
        start, end = start * f, end * f

        bits = message.decoded_bits_str[start-message.alignment_offset:end-message.alignment_offset]
        hexs = "".join(("{0:x}".format(int(bits[i:i + 4], 2)) for i in range(0, len(bits), 4)))
        decimals = str(int(bits, 2)) if len(bits) > 0 else ""

        self.ui.lBitsSelection.setText(bits)
        self.ui.lHexSelection.setText(hexs)
        self.ui.lDecimalSelection.setText(decimals)
        self.__set_decoding_error_label(message)

        self.ui.lblLabelValues.setText(self.tr("Labels for message #") + str(min_row + 1))
        if min_row != self.label_value_model.message_index:
            self.label_value_model.message_index = min_row

        active_group_ids = set()

        for group, tree_items in self.proto_tree_model.protocol_tree_items.items():
            for i, tree_item in enumerate(tree_items):
                proto = tree_item.protocol
                if proto.show and proto in self.rows_for_protocols:
                    if any(i in self.rows_for_protocols[proto] for i in range(min_row, max_row + 1)):
                        active_group_ids.add(group)
                        self.selected_protocols.add(proto)

        if active_group_ids != set(self.active_group_ids):
            self.active_group_ids = list(active_group_ids)
            self.active_group_ids.sort()

        if message.rssi > 0:
            self.ui.lblRSSI.setText(locale.format_string("%.2f dBm", 10*math.log10(message.rssi)))
        else:
            self.ui.lblRSSI.setText("-\u221e dBm")
        if isinstance(message.absolute_time, str):
            # For protocol files the abs time is the timestamp as string
            abs_time = message.absolute_time
        else:
            abs_time = Formatter.science_time(message.absolute_time)

        rel_time = Formatter.science_time(message.relative_time)
        self.ui.lTime.setText("{0} (+{1})".format(abs_time, rel_time))

        # Set Decoding Combobox
        self.ui.cbDecoding.blockSignals(True)
        different_encodings = any(msg.decoder != message.decoder for msg in selected_messages)
        self.ui.cbDecoding.setCurrentText("..." if different_encodings else message.decoder.name)
        self.ui.cbDecoding.blockSignals(False)

        self.updateUI(ignore_table_model=True, resize_table=False)

    @pyqtSlot(int)
    def on_ref_index_changed(self, new_ref_index: int):
        if new_ref_index != -1 and self.protocol_model.row_count:
            hide_correction = 0
            for i in range(0, self.protocol_model.row_count):
                if self.ui.tblViewProtocol.isRowHidden((new_ref_index + i) % self.protocol_model.row_count):
                    hide_correction = 0
                else:
                    hide_correction = i
                    break

            self.protocol_model.refindex = (new_ref_index + hide_correction) % self.protocol_model.row_count

        self.set_show_only_status()

    @pyqtSlot(QModelIndex, QModelIndex)
    def on_tbl_view_message_current_row_changed(self, current: QModelIndex, previous: QModelIndex):
        row = current.row()
        if row == -1:
            return

        self.active_message_type = self.proto_analyzer.message_types[row]
        self.ui.lblLabelValues.setText("Labels of {}".format(self.active_message_type.name))
        self.label_value_model.show_label_values = False
        self.label_value_model.update()

    @pyqtSlot(MessageType, list)
    def on_message_type_selected(self, message_type: MessageType, selected_messages: list):
        for msg in selected_messages:
            msg.message_type = message_type
        self.active_message_type = message_type
        self.protocol_model.update()

    @pyqtSlot()
    def on_message_type_dialog_accepted(self):
        self.update_automatic_assigned_message_types()

    @pyqtSlot()
    def on_participant_edited(self):
        self.refresh_assigned_participants_ui()

    @pyqtSlot(str)
    def on_label_shown_link_activated(self, link: str):
        if link == "reset_filter":
            self.ui.lineEditSearch.clear()
            self.show_all_rows()

    @pyqtSlot(str)
    def on_label_clear_alignment_link_activated(self, link: str):
        if link == "reset_alignment":
            self.align_messages(pattern="")

    @pyqtSlot()
    def on_protocol_updated(self):
        self.set_shown_protocols()
        self.ui.tblViewProtocol.zero_hide_offsets.clear()

    @pyqtSlot()
    def on_message_type_table_model_updated(self):
        self.ui.tblViewMessageTypes.open_persistent_editor(column=1)

    @pyqtSlot(MessageType)
    def on_message_type_removed(self, message_type: MessageType):
        for msg in self.proto_analyzer.messages:
            if msg.message_type == message_type:
                msg.message_type = self.proto_analyzer.default_message_type
        self.message_type_table_model.update()
        self.active_message_type = self.proto_analyzer.default_message_type
        self.protocol_model.update()
        self.label_value_model.update()

    @pyqtSlot(ProtocolLabel)
    def on_label_color_changed(self, lbl: ProtocolLabel):
        self.protocol_model.update()

    @pyqtSlot(MessageType)
    def on_message_type_visibility_changed(self, message_type: MessageType):
        self.set_message_type_visibility(message_type)

    @pyqtSlot(str)
    def on_message_type_name_edited(self, new_name: str):
        if self.ui.lblLabelValues.text().startswith("Labels of"):
            self.ui.lblLabelValues.setText("Labels of {}".format(new_name))



# -*- coding: utf-8 -*-

#
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ModulationSettings(object):
    def setupUi(self, ModulationSettings):
        ModulationSettings.setObjectName("ModulationSettings")
        ModulationSettings.resize(821, 635)
        self.verticalLayout = QtWidgets.QVBoxLayout(ModulationSettings)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBoxSniffSettings = QtWidgets.QGroupBox(ModulationSettings)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.groupBoxSniffSettings.setFont(font)
        self.groupBoxSniffSettings.setStyleSheet("QGroupBox\n"
"{\n"
"border: none;\n"
"}\n"
"\n"
"QGroupBox::title {\n"
"    subcontrol-origin: margin;\n"
"}\n"
"QGroupBox::indicator:unchecked {\n"
" image: url(:/icons/icons/collapse.svg)\n"
"}\n"
"QGroupBox::indicator:checked {\n"
" image: url(:/icons/icons/uncollapse.svg)\n"
"}")
        self.groupBoxSniffSettings.setFlat(True)
        self.groupBoxSniffSettings.setCheckable(True)
        self.groupBoxSniffSettings.setObjectName("groupBoxSniffSettings")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBoxSniffSettings)
        self.gridLayout_3.setContentsMargins(-1, 15, -1, -1)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.frame = QtWidgets.QFrame(self.groupBoxSniffSettings)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame.setLineWidth(0)
        self.frame.setObjectName("frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.labelModulationProfile = QtWidgets.QLabel(self.frame)
        self.labelModulationProfile.setObjectName("labelModulationProfile")
        self.verticalLayout_2.addWidget(self.labelModulationProfile)
        self.comboBoxModulationProfiles = QtWidgets.QComboBox(self.frame)
        self.comboBoxModulationProfiles.setObjectName("comboBoxModulationProfiles")
        self.verticalLayout_2.addWidget(self.comboBoxModulationProfiles)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.labelBitLength = QtWidgets.QLabel(self.frame)
        self.labelBitLength.setObjectName("labelBitLength")
        self.gridLayout.addWidget(self.labelBitLength, 0, 3, 1, 1)
        self.labelParamOneValue = QtWidgets.QLabel(self.frame)
        self.labelParamOneValue.setObjectName("labelParamOneValue")
        self.gridLayout.addWidget(self.labelParamOneValue, 1, 7, 1, 1)
        self.labelCarrierFrequencyValue = QtWidgets.QLabel(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelCarrierFrequencyValue.sizePolicy().hasHeightForWidth())
        self.labelCarrierFrequencyValue.setSizePolicy(sizePolicy)
        self.labelCarrierFrequencyValue.setObjectName("labelCarrierFrequencyValue")
        self.gridLayout.addWidget(self.labelCarrierFrequencyValue, 0, 1, 1, 1)
        self.labelModulationType = QtWidgets.QLabel(self.frame)
        self.labelModulationType.setObjectName("labelModulationType")
        self.gridLayout.addWidget(self.labelModulationType, 1, 0, 1, 1)
        self.labelBitLengthValue = QtWidgets.QLabel(self.frame)
        self.labelBitLengthValue.setObjectName("labelBitLengthValue")
        self.gridLayout.addWidget(self.labelBitLengthValue, 0, 4, 1, 1)
        self.labelParamOne = QtWidgets.QLabel(self.frame)
        self.labelParamOne.setObjectName("labelParamOne")
        self.gridLayout.addWidget(self.labelParamOne, 1, 6, 1, 1)
        self.labelModulationTypeValue = QtWidgets.QLabel(self.frame)
        self.labelModulationTypeValue.setObjectName("labelModulationTypeValue")
        self.gridLayout.addWidget(self.labelModulationTypeValue, 1, 1, 1, 1)
        self.line = QtWidgets.QFrame(self.frame)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 0, 2, 2, 1)
        self.labelCarrierFrequency = QtWidgets.QLabel(self.frame)
        self.labelCarrierFrequency.setObjectName("labelCarrierFrequency")
        self.gridLayout.addWidget(self.labelCarrierFrequency, 0, 0, 1, 1)
        self.line_2 = QtWidgets.QFrame(self.frame)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 0, 5, 2, 1)
        self.labelParamZero = QtWidgets.QLabel(self.frame)
        self.labelParamZero.setObjectName("labelParamZero")
        self.gridLayout.addWidget(self.labelParamZero, 0, 6, 1, 1)
        self.labelParamZeroValue = QtWidgets.QLabel(self.frame)
        self.labelParamZeroValue.setObjectName("labelParamZeroValue")
        self.gridLayout.addWidget(self.labelParamZeroValue, 0, 7, 1, 1)
        self.labelSampleRate = QtWidgets.QLabel(self.frame)
        self.labelSampleRate.setObjectName("labelSampleRate")
        self.gridLayout.addWidget(self.labelSampleRate, 1, 3, 1, 1)
        self.labelSampleRateValue = QtWidgets.QLabel(self.frame)
        self.labelSampleRateValue.setObjectName("labelSampleRateValue")
        self.gridLayout.addWidget(self.labelSampleRateValue, 1, 4, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.btnConfigurationDialog = QtWidgets.QPushButton(self.frame)
        icon = QtGui.QIcon.fromTheme("configure")
        self.btnConfigurationDialog.setIcon(icon)
        self.btnConfigurationDialog.setObjectName("btnConfigurationDialog")
        self.verticalLayout_2.addWidget(self.btnConfigurationDialog)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.gridLayout_3.addWidget(self.frame, 0, 0, 1, 1)
        self.verticalLayout.addWidget(self.groupBoxSniffSettings)

        self.retranslateUi(ModulationSettings)
        self.groupBoxSniffSettings.toggled['bool'].connect(self.frame.setVisible)
        ModulationSettings.setTabOrder(self.groupBoxSniffSettings, self.comboBoxModulationProfiles)
        ModulationSettings.setTabOrder(self.comboBoxModulationProfiles, self.btnConfigurationDialog)

    def retranslateUi(self, ModulationSettings):
        _translate = QtCore.QCoreApplication.translate
        ModulationSettings.setWindowTitle(_translate("ModulationSettings", "Form"))
        self.groupBoxSniffSettings.setTitle(_translate("ModulationSettings", "Modulation settings"))
        self.labelModulationProfile.setText(_translate("ModulationSettings", "Choose profile:"))
        self.labelBitLength.setText(_translate("ModulationSettings", "Bit Length:"))
        self.labelParamOneValue.setText(_translate("ModulationSettings", "TextLabel"))
        self.labelCarrierFrequencyValue.setText(_translate("ModulationSettings", "TextLabel"))
        self.labelModulationType.setText(_translate("ModulationSettings", "Modulation type:"))
        self.labelBitLengthValue.setText(_translate("ModulationSettings", "TextLabel"))
        self.labelParamOne.setText(_translate("ModulationSettings", "Param for 1:"))
        self.labelModulationTypeValue.setText(_translate("ModulationSettings", "TextLabel"))
        self.labelCarrierFrequency.setText(_translate("ModulationSettings", "Carrier Frequency:"))
        self.labelParamZero.setText(_translate("ModulationSettings", "Param for 0:"))
        self.labelParamZeroValue.setText(_translate("ModulationSettings", "TextLabel"))
        self.labelSampleRate.setText(_translate("ModulationSettings", "Sample Rate:"))
        self.labelSampleRateValue.setText(_translate("ModulationSettings", "TextLabel"))
        self.btnConfigurationDialog.setText(_translate("ModulationSettings", "Open modulation configuration dialog..."))

from . import urh_rc



from PyQt5.QtCore import QModelIndex, QAbstractItemModel, Qt
from PyQt5.QtWidgets import QStyledItemDelegate, QWidget, QStyleOptionViewItem, QLineEdit, QHBoxLayout, \
    QCompleter, QLabel, QSpinBox, QDirModel

from urh.ui.ExpressionLineEdit import ExpressionLineEdit
from urh.ui.RuleExpressionValidator import RuleExpressionValidator


class ExternalProgramWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        completer = QCompleter()
        completer.setModel(QDirModel(completer))
        self.line_edit_external_program = QLineEdit()
        self.line_edit_external_program.setCompleter(completer)
        self.line_edit_external_program.setPlaceholderText("Type in a path to external program.")

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout.addWidget(self.line_edit_external_program)

        self.setLayout(self.layout)


class RandomValueWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setAutoFillBackground(True)

        self.lbl_random_min = QLabel("Minimum (Decimal):")
        self.lbl_random_max = QLabel("Maximum (Decimal):")
        self.spinbox_random_min = QSpinBox()
        self.spinbox_random_max = QSpinBox()

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(10)
        self.layout.addWidget(self.lbl_random_min)
        self.layout.addWidget(self.spinbox_random_min)
        self.layout.addWidget(self.lbl_random_max)
        self.layout.addWidget(self.spinbox_random_max)

        self.spinbox_random_max.valueChanged.connect(self.on_max_value_changed)

        self.setLayout(self.layout)

    def on_max_value_changed(self, value):
        self.spinbox_random_min.setMaximum(value - 1)


class ProtocolValueDelegate(QStyledItemDelegate):
    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.controller = controller

    def createEditor(self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex):
        model = index.model()
        row = index.row()
        lbl = model.message_type[row]

        if lbl.value_type_index == 2:
            line_edit = ExpressionLineEdit(parent)
            line_edit.setPlaceholderText("(item1.length + 3) ^ 0x12")
            line_edit.setCompleter(QCompleter(self.controller.completer_model, line_edit))
            line_edit.setValidator(RuleExpressionValidator(self.controller.sim_expression_parser))
            line_edit.setToolTip(self.controller.sim_expression_parser.formula_help)
            return line_edit
        elif lbl.value_type_index == 3:
            return ExternalProgramWidget(parent)
        elif lbl.value_type_index == 4:
            random_widget = RandomValueWidget(parent)
            random_widget.spinbox_random_min.setMaximum(lbl.fuzz_maximum - 2)
            random_widget.spinbox_random_max.setMinimum(1)
            random_widget.spinbox_random_max.setMaximum(lbl.fuzz_maximum - 1)

            return random_widget
        else:
            return super().createEditor(parent, option, index)

    def setEditorData(self, editor: QWidget, index: QModelIndex):
        if isinstance(editor, ExternalProgramWidget):
            item = index.model().data(index)
            editor.line_edit_external_program.setText(item)
        elif isinstance(editor, RandomValueWidget):
            items = index.model().data(index, Qt.EditRole)
            editor.spinbox_random_max.setValue(items[1])
            editor.spinbox_random_min.setValue(items[0])
        else:
            super().setEditorData(editor, index)

    def setModelData(self, editor: QWidget, model: QAbstractItemModel, index: QModelIndex):
        if isinstance(editor, ExternalProgramWidget):
            model.setData(index, editor.line_edit_external_program.text(), Qt.EditRole)
        elif isinstance(editor, RandomValueWidget):
            model.setData(index, [editor.spinbox_random_min.value(), editor.spinbox_random_max.value()], Qt.EditRole)
        else:
            super().setModelData(editor, model, index)



from collections import OrderedDict
from multiprocessing import Array

import numpy as np

from urh.dev.native.Device import Device
from urh.dev.native.lib import bladerf
from multiprocessing.connection import Connection


class BladeRF(Device):
    SYNC_RX_CHUNK_SIZE = 16384
    SYNC_TX_CHUNK_SIZE = 16384

    DEVICE_LIB = bladerf
    ASYNCHRONOUS = False
    DEVICE_METHODS = Device.DEVICE_METHODS.copy()
    DEVICE_METHODS.update({
        Device.Command.SET_RF_GAIN.name: "set_gain",
        Device.Command.SET_CHANNEL_INDEX.name: "set_channel"
    })

    @classmethod
    def get_device_list(cls):
        return bladerf.get_device_list()

    @classmethod
    def adapt_num_read_samples_to_sample_rate(cls, sample_rate):
        cls.SYNC_RX_CHUNK_SIZE = 16384 * int(sample_rate / 1e6)

    @classmethod
    def setup_device(cls, ctrl_connection: Connection, device_identifier):
        if not device_identifier:
            device_identifier = ""

        ret = bladerf.open(device_identifier)
        if not device_identifier:
            ctrl_connection.send("OPEN:" + str(ret))
        else:
            ctrl_connection.send("OPEN ({}):{}".format(device_identifier, ret))

        ctrl_connection.send("If you experience problems, make sure you place a rbf file matching your device"
                             " at the correct location. See http://www.nuand.com/fpga_images/"
                             " and https://github.com/Nuand/bladeRF/wiki/FPGA-Autoloading")

        return ret == 0

    @classmethod
    def init_device(cls, ctrl_connection: Connection, is_tx: bool, parameters: OrderedDict) -> bool:
        bladerf.set_tx(is_tx)
        return super().init_device(ctrl_connection, is_tx, parameters)

    @classmethod
    def shutdown_device(cls, ctrl_connection, is_tx: bool):
        ret = bladerf.close()
        ctrl_connection.send("CLOSE:" + str(ret))
        return True

    @classmethod
    def prepare_sync_receive(cls, ctrl_connection: Connection):
        ctrl_connection.send("Initializing BladeRF..")
        ret = bladerf.prepare_sync()
        return ret

    @classmethod
    def receive_sync(cls, data_conn: Connection):
        bladerf.receive_sync(data_conn, cls.SYNC_RX_CHUNK_SIZE)

    @classmethod
    def prepare_sync_send(cls, ctrl_connection: Connection):
        ctrl_connection.send("Initializing BladeRF...")
        ret = bladerf.prepare_sync()
        return ret

    @classmethod
    def send_sync(cls, data):
        bladerf.send_sync(data)

    def __init__(self, center_freq, sample_rate, bandwidth, gain, if_gain=1, baseband_gain=1,
                 resume_on_full_receive_buffer=False):
        super().__init__(center_freq=center_freq, sample_rate=sample_rate, bandwidth=bandwidth,
                         gain=gain, if_gain=if_gain, baseband_gain=baseband_gain,
                         resume_on_full_receive_buffer=resume_on_full_receive_buffer)
        self.success = 0

    @property
    def has_multi_device_support(self):
        return True

    @property
    def device_parameters(self):
        return OrderedDict([(self.Command.SET_CHANNEL_INDEX.name, self.channel_index),
                            (self.Command.SET_FREQUENCY.name, self.frequency),
                            (self.Command.SET_SAMPLE_RATE.name, self.sample_rate),
                            (self.Command.SET_BANDWIDTH.name, self.bandwidth),
                            (self.Command.SET_RF_GAIN.name, self.gain),
                            ("identifier", self.device_serial)])

    @staticmethod
    def unpack_complex(buffer):
        unpacked = np.frombuffer(buffer, dtype=np.int16)
        result = np.empty(len(unpacked)//2, dtype=np.complex64)
        result.real = unpacked[::2] / 2048
        result.imag = unpacked[1::2] / 2048
        return result

    @staticmethod
    def pack_complex(complex_samples: np.ndarray):
        arr = Array("h", 2 * len(complex_samples), lock=False)
        numpy_view = np.frombuffer(arr, dtype=np.int16)
        numpy_view[:] = (2048 * complex_samples.view(np.float32)).astype(np.int16)
        return arr



import array
import copy
import itertools
from enum import Enum

import numpy

from urh.models.ProtocolTreeItem import ProtocolTreeItem
from urh.signalprocessing.Message import Message
from urh.signalprocessing.Modulator import Modulator
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.ProtocolGroup import ProtocolGroup
from urh.util.Logger import logger


class FuzzMode(Enum):
    successive = 0
    concurrent = 1
    exhaustive = 2


class ProtocolAnalyzerContainer(ProtocolAnalyzer):
    """
    A container to manage several ProtocolAnalyzers.
    This class is used by the generator to manage distinct protocols
    """

    def __init__(self):
        """

        :type modulators: list of Modulator
        """
        super().__init__(None)

        self.fuzz_pause = 10000

        self.__group = ProtocolGroup("GeneratorGroup")
        self.__group.add_protocol_item(ProtocolTreeItem(self, None))  # Warning: parent is None

    @property
    def protocol_labels(self):
        result = list(set(lbl for msg in self.messages for lbl in msg.message_type))
        result.sort()
        return result

    @property
    def multiple_fuzz_labels_per_message(self):
        return any(len(msg.active_fuzzing_labels) > 1 for msg in self.messages)

    def insert_protocol_analyzer(self, index: int, proto_analyzer: ProtocolAnalyzer):
        for msg in reversed(proto_analyzer.messages):
            self.messages.insert(index, Message(plain_bits=msg.decoded_bits, pause=msg.pause,
                                                message_type=copy.copy(msg.message_type),
                                                rssi=msg.rssi, modulator_index=0, decoder=msg.decoder,
                                                bit_len=msg.bit_len, participant=msg.participant))
        if len(self.pauses) > 0:
            self.fuzz_pause = self.pauses[0]

    def duplicate_lines(self, rows: list):
        for row in reversed(rows):
            try:
                self.messages.insert(max(rows) + 1, copy.deepcopy(self.messages[row]))
            except Exception as e:
                logger.error("Duplicating line ", str(e))
        self.qt_signals.line_duplicated.emit()

    def fuzz(self, mode: FuzzMode, default_pause=None):
        result = []
        appd_result = result.append

        added_message_indices = []

        for i, msg in enumerate(self.messages):
            labels = msg.active_fuzzing_labels
            appd_result(msg)

            if mode == FuzzMode.successive:
                combinations = [[(l.start, l.end, fuzz_val)] for l in labels for fuzz_val in l.fuzz_values[1:]]
            elif mode == FuzzMode.concurrent:
                num_values = numpy.max([len(l.fuzz_values) for l in labels]) if labels else 0
                f = lambda index, label: index if index < len(label.fuzz_values) else 0
                combinations = [[(l.start, l.end, l.fuzz_values[f(j, l)]) for l in labels] for j in
                                range(1, num_values)]
            elif mode == FuzzMode.exhaustive:
                pool = [[(l.start, l.end, fv) for fv in l.fuzz_values[1:]] for l in labels]
                combinations = list(itertools.product(*pool)) if labels else []
            else:
                raise ValueError("Unknown fuzz mode")

            self.qt_signals.fuzzing_started.emit(len(combinations))

            message_type = copy.copy(msg.message_type)
            for lbl in labels:
                lbl = copy.copy(lbl)
                lbl.fuzz_values = []
                lbl.fuzz_created = True
                message_type[message_type.index(lbl)] = lbl

            for j, combination in enumerate(combinations):
                cpy_bits = msg.plain_bits[:]
                for start, end, fuz_val in combination:
                    cpy_bits[start:end] = array.array("B", map(int, fuz_val))

                pause = default_pause if default_pause is not None else msg.pause
                fuz_msg = Message(plain_bits=cpy_bits, pause=pause,
                                  rssi=msg.rssi, message_type=message_type,
                                  modulator_index=msg.modulator_index,
                                  decoder=msg.decoder, fuzz_created=True, participant=msg.participant)
                added_message_indices.append(i + j + 1)
                appd_result(fuz_msg)
                if j % 10000 == 0:
                    self.qt_signals.current_fuzzing_message_changed.emit(j)

        self.qt_signals.fuzzing_finished.emit()
        self.messages = result  # type: list[Message]
        return added_message_indices

    def fuzz_successive(self, default_pause=None):
        """
        Führt ein sukzessives Fuzzing über alle aktiven Fuzzing Label durch.
        Sequentiell heißt, ein Label wird durchgefuzzt und alle anderen Labels bleiben auf Standardwert.
        Das entspricht dem Vorgang nacheinander immer nur ein Label aktiv zu setzen.
        """
        return self.fuzz(FuzzMode.successive, default_pause=default_pause)

    def fuzz_concurrent(self, default_pause=None):
        """
        Führt ein gleichzeitiges Fuzzing durch, das heißt bei mehreren Labels pro Message werden alle Labels
        gleichzeitig iteriert. Wenn ein Label keine FuzzValues mehr übrig hat,
        wird der erste Fuzzing Value (per Definition der Standardwert) genommen.
        """
        return self.fuzz(FuzzMode.concurrent, default_pause=default_pause)

    def fuzz_exhaustive(self, default_pause=None):
        """
        Führt ein vollständiges Fuzzing durch. D.h. wenn es mehrere Label pro Message gibt, werden alle
        möglichen Kombinationen erzeugt (Kreuzprodukt!)
        """
        return self.fuzz(FuzzMode.exhaustive, default_pause=default_pause)

    def create_fuzzing_label(self, start, end, msg_index) -> ProtocolLabel:
        fuz_lbl = self.messages[msg_index].message_type.add_protocol_label(start=start, end=end)
        return fuz_lbl

    def set_decoder_for_messages(self, decoder, messages=None):
        raise NotImplementedError("Encoding cant be set in Generator!")

    def to_xml_file(self, filename: str, decoders, participants, tag_name="fuzz_profile",
                    include_message_types=True, write_bits=True, modulators=None):
        super().to_xml_file(filename=filename, decoders=decoders, participants=participants, tag_name=tag_name,
                            include_message_types=include_message_types, write_bits=write_bits, modulators=modulators)

    def from_xml_file(self, filename: str, read_bits=True):
        super().from_xml_file(filename=filename, read_bits=read_bits)

    def clear(self):
        self.messages[:] = []



from collections import OrderedDict

import numpy as np

from multiprocessing.connection import Connection
from urh.dev.native.Device import Device

from urh.dev.native.lib import rtlsdr
from urh.util.Logger import logger


class RTLSDR(Device):
    DEVICE_LIB = rtlsdr
    ASYNCHRONOUS = False
    DEVICE_METHODS = Device.DEVICE_METHODS.copy()
    DEVICE_METHODS.update({
        Device.Command.SET_RF_GAIN.name: "set_tuner_gain",
        Device.Command.SET_RF_GAIN.name+"_get_allowed_values": "get_tuner_gains",
        Device.Command.SET_BANDWIDTH.name: "set_tuner_bandwidth",
        Device.Command.SET_FREQUENCY_CORRECTION.name: "set_freq_correction",
        Device.Command.SET_DIRECT_SAMPLING_MODE.name: "set_direct_sampling"
    })

    @classmethod
    def get_device_list(cls):
        return rtlsdr.get_device_list()

    @classmethod
    def setup_device(cls, ctrl_connection: Connection, device_identifier):
        # identifier gets set in self.receive_process_arguments
        device_number = int(device_identifier)
        ret = rtlsdr.open(device_number)
        ctrl_connection.send("OPEN (#{}):{}".format(device_number, ret))
        return ret == 0

    @classmethod
    def prepare_sync_receive(cls, ctrl_connection: Connection):
        ret = rtlsdr.reset_buffer()
        ctrl_connection.send("RESET_BUFFER:" + str(ret))
        return ret

    @classmethod
    def receive_sync(cls, data_conn: Connection):
        data_conn.send_bytes(rtlsdr.read_sync())

    @classmethod
    def shutdown_device(cls, ctrl_connection, is_tx: bool):
        logger.debug("RTLSDR: closing device")
        ret = rtlsdr.close()
        ctrl_connection.send("CLOSE:" + str(ret))

    def __init__(self, freq, gain, srate, device_number, resume_on_full_receive_buffer=False):
        super().__init__(center_freq=freq, sample_rate=srate, bandwidth=0,
                         gain=gain, if_gain=1, baseband_gain=1,
                         resume_on_full_receive_buffer=resume_on_full_receive_buffer)

        self.success = 0
        self.bandwidth_is_adjustable = self.get_bandwidth_is_adjustable()  # e.g. not in Manjaro Linux / Ubuntu 14.04

        self.device_number = device_number

        self.error_codes = {
            -100: "Method not available in installed driver."
        }

    @staticmethod
    def get_bandwidth_is_adjustable():
        return rtlsdr.bandwidth_is_adjustable()

    @property
    def device_parameters(self):
        return OrderedDict([(self.Command.SET_FREQUENCY.name, self.frequency),
                            (self.Command.SET_SAMPLE_RATE.name, self.sample_rate),
                            (self.Command.SET_BANDWIDTH.name, self.bandwidth),
                            (self.Command.SET_FREQUENCY_CORRECTION.name, self.freq_correction),
                            (self.Command.SET_DIRECT_SAMPLING_MODE.name, self.direct_sampling_mode),
                            (self.Command.SET_RF_GAIN.name, self.gain),
                            ("identifier", self.device_number)])

    @property
    def has_multi_device_support(self):
        return True

    def set_device_bandwidth(self, bandwidth):
        if self.bandwidth_is_adjustable:
            super().set_device_bandwidth(bandwidth)
        else:
            logger.warning("Setting the bandwidth is not supported by your RTL-SDR driver version.")

    @staticmethod
    def unpack_complex(buffer):
        """
        The raw, captured IQ data is 8 bit unsigned data.

        :return:
        """
        unpacked = np.frombuffer(buffer, dtype=[('r', np.uint8), ('i', np.uint8)])
        result = np.empty(len(unpacked), dtype=np.complex64)
        result.real = (unpacked['r'] / 127.5) - 1.0
        result.imag = (unpacked['i'] / 127.5) - 1.0
        return result




import array
import os
import shlex
import shutil
import subprocess
import sys
import time
from xml.dom import minidom
from xml.etree import ElementTree as ET

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFontDatabase, QFont
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QSplitter
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QPlainTextEdit, QTableWidgetItem
from urh import constants
from urh.util.Logger import logger

PROJECT_PATH = None   # for referencing in external program calls

BCD_ERROR_SYMBOL = "?"
BCD_LUT = {"{0:04b}".format(i): str(i) if i < 10 else BCD_ERROR_SYMBOL for i in range(16)}
BCD_REVERSE_LUT = {str(i): "{0:04b}".format(i) for i in range(10)}
BCD_REVERSE_LUT[BCD_ERROR_SYMBOL] = "0000"

DEFAULT_PROGRAMS_WINDOWS = {}


def profile(func):
    def func_wrapper(*args):
        t = time.perf_counter()
        result = func(*args)
        print("{} took {:.2f}ms".format(func, 1000 * (time.perf_counter() - t)))
        return result

    return func_wrapper


def set_icon_theme():
    if sys.platform != "linux" or constants.SETTINGS.value("icon_theme_index", 0, int) == 0:
        # noinspection PyUnresolvedReferences
        import urh.ui.xtra_icons_rc
        QIcon.setThemeName("oxy")
    else:
        QIcon.setThemeName("")


def set_shared_library_path():
    shared_lib_dir = get_shared_library_path()

    if shared_lib_dir:

        if sys.platform == "win32":
            current_path =  os.environ.get("PATH", '')
            if not current_path.startswith(shared_lib_dir):
                os.environ["PATH"] = shared_lib_dir + os.pathsep + current_path
        else:
            # LD_LIBRARY_PATH will not be considered at runtime so we explicitly load the .so's we need
            exts = [".so"] if sys.platform == "linux" else [".so", ".dylib"]
            import ctypes

            for lib in sorted(os.listdir(shared_lib_dir)):
                if any(lib.endswith(ext) for ext in exts):
                    lib_path = os.path.join(shared_lib_dir, lib)
                    if os.path.isfile(lib_path):
                        try:
                            ctypes.cdll.LoadLibrary(lib_path)
                        except Exception as e:
                            logger.exception(e)


def get_shared_library_path():
    if hasattr(sys, "frozen"):
        return os.path.dirname(sys.executable)

    util_dir = os.path.dirname(os.path.realpath(__file__)) if not os.path.islink(__file__) \
        else os.path.dirname(os.path.realpath(os.readlink(__file__)))
    urh_dir = os.path.realpath(os.path.join(util_dir, ".."))
    assert os.path.isdir(urh_dir)

    shared_lib_dir = os.path.realpath(os.path.join(urh_dir, "dev", "native", "lib", "shared"))
    if os.path.isdir(shared_lib_dir):
        return shared_lib_dir
    else:
        return ""


def convert_bits_to_string(bits, output_view_type: int, pad_zeros=False, lsb=False, lsd=False, endianness="big"):
    """
    Convert bit array to string
    :param endianness: Endianness little or big
    :param bits: Bit array
    :param output_view_type: Output view type index
    0 = bit, 1=hex, 2=ascii, 3=decimal 4=binary coded decimal (bcd)
    :param pad_zeros:
    :param lsb: Least Significant Bit   -> Reverse bits first
    :param lsd: Least Significant Digit -> Reverse result at end
    :return:
    """
    bits_str = "".join(["1" if b else "0" for b in bits])

    if output_view_type == 4:
        # For BCD we need to enforce padding
        pad_zeros = True

    if pad_zeros and output_view_type in (1, 2, 4):
        n = 4 if output_view_type in (1, 4) else 8 if output_view_type == 2 else 1
        bits_str += "0" * ((n - (len(bits_str) % n)) % n)

    if lsb:
        # Reverse bit string
        bits_str = bits_str[::-1]

    if endianness == "little":
        # reverse byte wise
        bits_str = "".join(bits_str[max(i - 8, 0):i] for i in range(len(bits_str), 0, -8))

    if output_view_type == 0:  # bit
        result = bits_str

    elif output_view_type == 1:  # hex
        result = "".join(["{0:x}".format(int(bits_str[i:i + 4], 2)) for i in range(0, len(bits_str), 4)])

    elif output_view_type == 2:  # ascii
        result = "".join(map(chr,
                             [int("".join(bits_str[i:i + 8]), 2) for i in range(0, len(bits_str), 8)]))

    elif output_view_type == 3:  # decimal
        try:
            result = str(int(bits_str, 2))
        except ValueError:
            return None
    elif output_view_type == 4:  # bcd
        result = "".join([BCD_LUT[bits_str[i:i + 4]] for i in range(0, len(bits_str), 4)])
    else:
        raise ValueError("Unknown view type")

    if lsd:
        # reverse result
        return result[::-1]
    else:
        return result


def hex2bit(hex_str: str) -> array.array:
    if not isinstance(hex_str, str):
        return array.array("B", [])

    if hex_str[:2] == "0x":
        hex_str = hex_str[2:]

    try:
        bitstring = "".join("{0:04b}".format(int(h, 16)) for h in hex_str)
        return array.array("B", [True if x == "1" else False for x in bitstring])
    except (TypeError, ValueError) as e:
        logger.error(e)
        result = array.array("B", [])

    return result


def ascii2bit(ascii_str: str) -> array.array:
    if not isinstance(ascii_str, str):
        return array.array("B", [])

    try:
        bitstring = "".join("{0:08b}".format(ord(c)) for c in ascii_str)
        return array.array("B", [True if x == "1" else False for x in bitstring])
    except (TypeError, ValueError) as e:
        logger.error(e)
        result = array.array("B", [])

    return result


def decimal2bit(number: str, num_bits: int) -> array.array:
    try:
        number = int(number)
    except ValueError as e:
        logger.error(e)
        return array.array("B", [])

    fmt_str = "{0:0" + str(num_bits) + "b}"
    return array.array("B", map(int, fmt_str.format(number)))


def bcd2bit(value: str) -> array.array:
    try:
        return array.array("B", map(int, "".join(BCD_REVERSE_LUT[c] for c in value)))
    except Exception as e:
        logger.error(e)
        return array.array("B", [])


def convert_string_to_bits(value: str, display_format: int, target_num_bits: int) -> array.array:
    if display_format == 0:
        result = string2bits(value)
    elif display_format == 1:
        result = hex2bit(value)
    elif display_format == 2:
        result = ascii2bit(value)
    elif display_format == 3:
        result = decimal2bit(value, target_num_bits)
    elif display_format == 4:
        result = bcd2bit(value)
    else:
        raise ValueError("Unknown display format {}".format(display_format))

    if len(result) == 0:
        raise ValueError("Error during conversion.")

    if len(result) < target_num_bits:
        # pad with zeros
        return result + array.array("B", [0] * (target_num_bits-len(result)))
    else:
        return result[:target_num_bits]


def create_textbox_dialog(content: str, title: str, parent) -> QDialog:
    d = QDialog(parent)
    d.resize(800, 600)
    d.setWindowTitle(title)
    layout = QVBoxLayout(d)
    text_edit = QPlainTextEdit(content)
    text_edit.setReadOnly(True)
    layout.addWidget(text_edit)
    d.setLayout(layout)
    return d


def string2bits(bit_str: str) -> array.array:
    return array.array("B", map(int, bit_str))


def bit2hex(bits: array.array, pad_zeros=False) -> str:
    return convert_bits_to_string(bits, 1, pad_zeros)


def number_to_bits(n: int, length: int) -> array.array:
    fmt = "{0:0" + str(length) + "b}"
    return array.array("B", map(int, fmt.format(n)))


def aggregate_bits(bits: array.array, size=4) -> array.array:
    result = array.array("B", [])

    for i in range(0, len(bits), size):
        h = 0
        for k in range(size):
            try:
                h += (2 ** (size - 1 - k)) * bits[i + k]
            except IndexError:
                # Implicit padding with zeros
                continue
        result.append(h)

    return result


def clip(value, minimum, maximum):
    return max(minimum, min(value, maximum))


def file_can_be_opened(filename: str):
    try:
        open(filename, "r").close()
        return True
    except Exception as e:
        if not isinstance(e, FileNotFoundError):
            logger.debug(str(e))
        return False


def create_table_item(content):
    item = QTableWidgetItem(str(content))
    item.setFlags(item.flags() & ~Qt.ItemIsEditable)
    return item


def write_xml_to_file(xml_tag: ET.Element, filename: str):
    xml_str = minidom.parseString(ET.tostring(xml_tag)).toprettyxml(indent="  ")
    with open(filename, "w") as f:
        for line in xml_str.split("\n"):
            if line.strip():
                f.write(line + "\n")


def get_monospace_font() -> QFont:
    fixed_font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
    fixed_font.setPointSize(QApplication.instance().font().pointSize())
    return fixed_font


def get_name_from_filename(filename: str):
    if not isinstance(filename, str):
        return "No Name"

    return os.path.basename(filename).split(".")[0]


def get_default_windows_program_for_extension(extension: str):
    if os.name != "nt":
        return None

    if not extension.startswith("."):
        extension = "." + extension

    if extension in DEFAULT_PROGRAMS_WINDOWS:
        return DEFAULT_PROGRAMS_WINDOWS[extension]

    try:
        assoc = subprocess.check_output("assoc " + extension, shell=True, stderr=subprocess.PIPE).decode().split("=")[1]
        ftype = subprocess.check_output("ftype " + assoc, shell=True).decode().split("=")[1].split(" ")[0]
        ftype = ftype.replace('"', '')
        assert shutil.which(ftype) is not None
    except Exception:
        return None

    DEFAULT_PROGRAMS_WINDOWS[extension] = ftype
    return ftype


def parse_command(command: str):
    try:
        posix = os.name != "nt"
        splitted = shlex.split(command, posix=posix)
        # strip quotations
        if not posix:
            splitted = [s.replace('"', '').replace("'", "") for s in splitted]
    except ValueError:
        splitted = []  # e.g. when missing matching "

    if len(splitted) == 0:
        return "", []

    cmd = splitted.pop(0)
    if PROJECT_PATH is not None and not os.path.isabs(cmd) and shutil.which(cmd) is None:
        # Path relative to project path
        cmd = os.path.normpath(os.path.join(PROJECT_PATH, cmd))
    cmd = [cmd]

    # This is for legacy support, if you have filenames with spaces and did not quote them
    while shutil.which(" ".join(cmd)) is None and len(splitted) > 0:
        cmd.append(splitted.pop(0))

    return " ".join(cmd), splitted


def run_command(command, param: str = None, use_stdin=False, detailed_output=False, return_rc=False):
    cmd, arg = parse_command(command)
    if shutil.which(cmd) is None:
        logger.error("Could not find {}".format(cmd))
        return ""

    startupinfo = None
    if os.name == 'nt':
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        if "." in cmd:
            default_app = get_default_windows_program_for_extension(cmd.split(".")[-1])
            if default_app:
                arg.insert(0, cmd)
                cmd = default_app

    call_list = [cmd] + arg
    try:
        if detailed_output:
            if param is not None:
                call_list.append(param)

            p = subprocess.Popen(call_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE, startupinfo=startupinfo)
            out, err = p.communicate()
            result = "{} exited with {}".format(" ".join(call_list), p.returncode)
            if out.decode():
                result += " stdout: {}".format(out.decode())
            if err.decode():
                result += " stderr: {}".format(err.decode())

            if return_rc:
                return result, p.returncode
            else:
                return result
        elif use_stdin:
            p = subprocess.Popen(call_list, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                 startupinfo=startupinfo)
            param = param.encode() if param is not None else None
            out, _ = p.communicate(param)
            if return_rc:
                return out.decode(), p.returncode
            else:
                return out.decode()
        else:
            if param is not None:
                call_list.append(param)

            if return_rc:
                raise ValueError("Return Code not supported for this configuration")

            return subprocess.check_output(call_list, stderr=subprocess.PIPE, startupinfo=startupinfo).decode()
    except Exception as e:
        msg = "Could not run {} ({})".format(cmd, e)
        logger.error(msg)
        if detailed_output:
            return msg
        else:
            return ""


def validate_command(command: str):
    if not isinstance(command, str):
        return False

    cmd, _ = parse_command(command)
    return shutil.which(cmd) is not None


def set_splitter_stylesheet(splitter: QSplitter):
    splitter.setHandleWidth(4)
    bgcolor = constants.BGCOLOR.lighter(150)
    r, g, b, a = bgcolor.red(), bgcolor.green(), bgcolor.blue(), bgcolor.alpha()
    splitter.setStyleSheet("QSplitter::handle:vertical {{margin: 4px 0px; "
                           "background-color: qlineargradient(x1:0, y1:0, x2:1, y2:0, "
                           "stop:0.2 rgba(255, 255, 255, 0),"
                           "stop:0.5 rgba({0}, {1}, {2}, {3}),"
                           "stop:0.8 rgba(255, 255, 255, 0));"
                           "image: url(:/icons/icons/splitter_handle_horizontal.svg);}}"
                           "QSplitter::handle:horizontal {{margin: 4px 0px; "
                           "background-color: qlineargradient(x1:0, y1:0, x2:0, y2:1, "
                           "stop:0.2 rgba(255, 255, 255, 0),"
                           "stop:0.5 rgba({0}, {1}, {2}, {3}),"
                           "stop:0.8 rgba(255, 255, 255, 0));"
                           "image: url(:/icons/icons/splitter_handle_vertical.svg);}}".format(r, g, b, a))



from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtGui import QBrush, QColor, QIcon, QPen
from PyQt5.QtWidgets import QMessageBox

from urh import constants
from urh.controller.dialogs.SendRecvDialog import SendRecvDialog
from urh.dev.VirtualDevice import VirtualDevice, Mode
from urh.signalprocessing.Signal import Signal
from urh.ui.painting.SignalSceneManager import SignalSceneManager
from urh.util import FileOperator
from urh.util.Logger import logger


class SendDialog(SendRecvDialog):
    def __init__(self, project_manager, modulated_data, modulation_msg_indices=None, continuous_send_mode=False,
                 parent=None, testing_mode=False):
        super().__init__(project_manager, is_tx=True, continuous_send_mode=continuous_send_mode,
                         parent=parent, testing_mode=testing_mode)

        self.graphics_view = self.ui.graphicsViewSend
        self.ui.stackedWidget.setCurrentWidget(self.ui.page_send)
        self.hide_receive_ui_items()

        self.ui.btnStart.setIcon(QIcon.fromTheme("media-playback-start"))
        self.setWindowTitle("Send Signal")
        self.setWindowIcon(QIcon.fromTheme("media-playback-start"))
        self.ui.btnStart.setToolTip("Send data")
        self.ui.btnStop.setToolTip("Stop sending")
        self.device_is_sending = False
        self.modulation_msg_indices = modulation_msg_indices

        if self.modulation_msg_indices is not None:
            self.ui.progressBarMessage.setMaximum(len(self.modulation_msg_indices))
        else:
            self.ui.progressBarMessage.hide()
            self.ui.labelCurrentMessage.hide()

        if modulated_data is not None:
            # modulated_data is none in continuous send mode
            self.ui.progressBarSample.setMaximum(len(modulated_data))
            samp_rate = self.device_settings_widget.ui.spinBoxSampleRate.value()
            signal = Signal.from_samples(modulated_data, "Modulated Preview", samp_rate)
            self.scene_manager = SignalSceneManager(signal, parent=self)
            self.send_indicator = self.scene_manager.scene.addRect(0, -2, 0, 4,
                                                                   QPen(QColor(Qt.transparent), 0),
                                                                   QBrush(constants.SEND_INDICATOR_COLOR))
            self.send_indicator.stackBefore(self.scene_manager.scene.selection_area)
            self.scene_manager.init_scene()
            self.graphics_view.set_signal(signal)
            self.graphics_view.sample_rate = samp_rate

            self.create_connects()
            self.device_settings_widget.update_for_new_device(overwrite_settings=False)

    def create_connects(self):
        super().create_connects()

        self.graphics_view.save_as_clicked.connect(self.on_graphics_view_save_as_clicked)
        self.scene_manager.signal.data_edited.connect(self.on_signal_data_edited)

    def _update_send_indicator(self, width: int):
        y, h = self.ui.graphicsViewSend.view_rect().y(), self.ui.graphicsViewSend.view_rect().height()
        self.send_indicator.setRect(0, y - h, width, 2 * h + abs(y))

    def set_current_message_progress_bar_value(self, current_sample: int):
        if self.modulation_msg_indices is not None:
            msg_index = next((i for i, sample in enumerate(self.modulation_msg_indices) if sample >= current_sample),
                             len(self.modulation_msg_indices))
            self.ui.progressBarMessage.setValue(msg_index + 1)

    def update_view(self):
        if super().update_view():
            self._update_send_indicator(self.device.current_index)
            self.ui.progressBarSample.setValue(self.device.current_index)
            self.set_current_message_progress_bar_value(self.device.current_index)

            if not self.device.sending_finished:
                self.ui.lblCurrentRepeatValue.setText(str(self.device.current_iteration + 1))
            else:
                self.ui.btnStop.click()
                self.ui.lblCurrentRepeatValue.setText("Sending finished")

    def init_device(self):
        device_name = self.selected_device_name
        num_repeats = self.device_settings_widget.ui.spinBoxNRepeat.value()
        sts = self.scene_manager.signal._fulldata

        self.device = VirtualDevice(self.backend_handler, device_name, Mode.send, samples_to_send=sts,
                                    device_ip="192.168.10.2", sending_repeats=num_repeats, parent=self)
        self._create_device_connects()

    @pyqtSlot()
    def on_graphics_view_save_as_clicked(self):
        filename = FileOperator.get_save_file_name("signal.complex")
        if filename:
            try:
                try:
                    self.scene_manager.signal.sample_rate = self.device.sample_rate
                except Exception as e:
                    logger.exception(e)

                self.scene_manager.signal.save_as(filename)
            except Exception as e:
                QMessageBox.critical(self, self.tr("Error saving signal"), e.args[0])

    @pyqtSlot()
    def on_signal_data_edited(self):
        signal = self.scene_manager.signal
        self.ui.progressBarSample.setMaximum(signal.num_samples)
        self.device.samples_to_send = signal.data
        self.scene_manager.init_scene()
        self.ui.graphicsViewSend.redraw_view()

    @pyqtSlot()
    def on_start_clicked(self):
        super().on_start_clicked()
        if self.ui.progressBarSample.value() >= self.ui.progressBarSample.maximum() - 1:
            self.on_clear_clicked()

        if self.device_is_sending:
            self.device.stop("Sending paused by user")
        else:
            self.device.start()

    @pyqtSlot()
    def on_stop_clicked(self):
        super().on_stop_clicked()
        self.on_clear_clicked()

    @pyqtSlot()
    def on_device_stopped(self):
        super().on_device_stopped()
        self.ui.btnStart.setIcon(QIcon.fromTheme("media-playback-start"))
        self.ui.btnStart.setText("Start")
        self.ui.btnStart.setToolTip("Start sending")
        self.device_is_sending = False

    @pyqtSlot()
    def on_device_started(self):
        super().on_device_started()
        self.device_is_sending = True
        self.ui.btnStart.setEnabled(True)
        self.ui.btnStart.setIcon(QIcon.fromTheme("media-playback-pause"))
        self.ui.btnStart.setText("Pause")
        self.set_device_ui_items_enabled(False)

    @pyqtSlot()
    def on_clear_clicked(self):
        self._update_send_indicator(0)
        self.reset()



# -*- coding: utf-8 -*-

#
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ProjectDialog(object):
    def setupUi(self, ProjectDialog):
        ProjectDialog.setObjectName("ProjectDialog")
        ProjectDialog.resize(803, 936)
        self.verticalLayout = QtWidgets.QVBoxLayout(ProjectDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lNewProject = QtWidgets.QLabel(ProjectDialog)
        font = QtGui.QFont()
        font.setPointSize(16)
        font.setBold(True)
        font.setWeight(75)
        self.lNewProject.setFont(font)
        self.lNewProject.setObjectName("lNewProject")
        self.verticalLayout.addWidget(self.lNewProject)
        self.lblName = QtWidgets.QLabel(ProjectDialog)
        self.lblName.setObjectName("lblName")
        self.verticalLayout.addWidget(self.lblName)
        spacerItem = QtWidgets.QSpacerItem(17, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_5 = QtWidgets.QLabel(ProjectDialog)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 3, 4, 1, 2)
        self.lineEdit_Path = QtWidgets.QLineEdit(ProjectDialog)
        self.lineEdit_Path.setStatusTip("")
        self.lineEdit_Path.setObjectName("lineEdit_Path")
        self.gridLayout.addWidget(self.lineEdit_Path, 0, 3, 1, 1)
        self.label_7 = QtWidgets.QLabel(ProjectDialog)
        self.label_7.setText("")
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 1, 0, 1, 2)
        self.lblNewPath = QtWidgets.QLabel(ProjectDialog)
        self.lblNewPath.setObjectName("lblNewPath")
        self.gridLayout.addWidget(self.lblNewPath, 1, 3, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 57, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 15, 4, 1, 2)
        self.txtEdDescription = QtWidgets.QPlainTextEdit(ProjectDialog)
        self.txtEdDescription.setObjectName("txtEdDescription")
        self.gridLayout.addWidget(self.txtEdDescription, 10, 3, 1, 1)
        self.tblParticipants = ParticipantTableView(ProjectDialog)
        self.tblParticipants.setAlternatingRowColors(True)
        self.tblParticipants.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.tblParticipants.setObjectName("tblParticipants")
        self.tblParticipants.horizontalHeader().setCascadingSectionResizes(False)
        self.tblParticipants.horizontalHeader().setDefaultSectionSize(100)
        self.tblParticipants.horizontalHeader().setStretchLastSection(True)
        self.tblParticipants.verticalHeader().setCascadingSectionResizes(True)
        self.tblParticipants.verticalHeader().setStretchLastSection(False)
        self.gridLayout.addWidget(self.tblParticipants, 11, 3, 5, 1)
        self.spinBoxBandwidth = KillerDoubleSpinBox(ProjectDialog)
        self.spinBoxBandwidth.setDecimals(3)
        self.spinBoxBandwidth.setMinimum(1.0)
        self.spinBoxBandwidth.setMaximum(999999999999.0)
        self.spinBoxBandwidth.setSingleStep(0.001)
        self.spinBoxBandwidth.setProperty("value", 1000000.0)
        self.spinBoxBandwidth.setObjectName("spinBoxBandwidth")
        self.gridLayout.addWidget(self.spinBoxBandwidth, 5, 3, 1, 1)
        self.label_3 = QtWidgets.QLabel(ProjectDialog)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 4, 0, 1, 2)
        self.spinBoxFreq = KillerDoubleSpinBox(ProjectDialog)
        self.spinBoxFreq.setDecimals(3)
        self.spinBoxFreq.setMinimum(0.01)
        self.spinBoxFreq.setMaximum(1000000000000.0)
        self.spinBoxFreq.setSingleStep(0.001)
        self.spinBoxFreq.setProperty("value", 433920000.0)
        self.spinBoxFreq.setObjectName("spinBoxFreq")
        self.gridLayout.addWidget(self.spinBoxFreq, 4, 3, 1, 1)
        self.spinBoxSampleRate = KillerDoubleSpinBox(ProjectDialog)
        self.spinBoxSampleRate.setDecimals(3)
        self.spinBoxSampleRate.setMinimum(0.01)
        self.spinBoxSampleRate.setMaximum(1000000000000.0)
        self.spinBoxSampleRate.setSingleStep(0.001)
        self.spinBoxSampleRate.setProperty("value", 1000000.0)
        self.spinBoxSampleRate.setObjectName("spinBoxSampleRate")
        self.gridLayout.addWidget(self.spinBoxSampleRate, 3, 3, 1, 1)
        self.lineEditBroadcastAddress = QtWidgets.QLineEdit(ProjectDialog)
        self.lineEditBroadcastAddress.setInputMask("")
        self.lineEditBroadcastAddress.setObjectName("lineEditBroadcastAddress")
        self.gridLayout.addWidget(self.lineEditBroadcastAddress, 16, 3, 1, 1)
        self.line = QtWidgets.QFrame(ProjectDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 9, 0, 1, 6)
        self.label_10 = QtWidgets.QLabel(ProjectDialog)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 5, 0, 1, 2)
        self.btnAddParticipant = QtWidgets.QToolButton(ProjectDialog)
        icon = QtGui.QIcon.fromTheme("list-add")
        self.btnAddParticipant.setIcon(icon)
        self.btnAddParticipant.setObjectName("btnAddParticipant")
        self.gridLayout.addWidget(self.btnAddParticipant, 11, 4, 1, 2)
        self.btnSelectPath = QtWidgets.QToolButton(ProjectDialog)
        self.btnSelectPath.setObjectName("btnSelectPath")
        self.gridLayout.addWidget(self.btnSelectPath, 0, 4, 1, 2)
        self.line_2 = QtWidgets.QFrame(ProjectDialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 2, 0, 1, 6)
        self.label_6 = QtWidgets.QLabel(ProjectDialog)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 4, 4, 1, 2)
        self.label_2 = QtWidgets.QLabel(ProjectDialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 3, 0, 1, 2)
        self.label = QtWidgets.QLabel(ProjectDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 2)
        self.btnRemoveParticipant = QtWidgets.QToolButton(ProjectDialog)
        icon = QtGui.QIcon.fromTheme("list-remove")
        self.btnRemoveParticipant.setIcon(icon)
        self.btnRemoveParticipant.setObjectName("btnRemoveParticipant")
        self.gridLayout.addWidget(self.btnRemoveParticipant, 12, 4, 1, 2)
        self.spinBoxGain = QtWidgets.QSpinBox(ProjectDialog)
        self.spinBoxGain.setMinimum(1)
        self.spinBoxGain.setProperty("value", 20)
        self.spinBoxGain.setObjectName("spinBoxGain")
        self.gridLayout.addWidget(self.spinBoxGain, 6, 3, 1, 1)
        self.label_4 = QtWidgets.QLabel(ProjectDialog)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 16, 0, 1, 2)
        self.label_8 = QtWidgets.QLabel(ProjectDialog)
        self.label_8.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 10, 0, 1, 2)
        self.label_12 = QtWidgets.QLabel(ProjectDialog)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 5, 4, 1, 2)
        self.label_11 = QtWidgets.QLabel(ProjectDialog)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 6, 0, 1, 2)
        self.btnUp = QtWidgets.QToolButton(ProjectDialog)
        icon = QtGui.QIcon.fromTheme("go-up")
        self.btnUp.setIcon(icon)
        self.btnUp.setObjectName("btnUp")
        self.gridLayout.addWidget(self.btnUp, 13, 4, 1, 1)
        self.lOpenSpectrumAnalyzer = QtWidgets.QLabel(ProjectDialog)
        self.lOpenSpectrumAnalyzer.setTextFormat(QtCore.Qt.AutoText)
        self.lOpenSpectrumAnalyzer.setOpenExternalLinks(False)
        self.lOpenSpectrumAnalyzer.setObjectName("lOpenSpectrumAnalyzer")
        self.gridLayout.addWidget(self.lOpenSpectrumAnalyzer, 8, 0, 1, 4)
        self.label_9 = QtWidgets.QLabel(ProjectDialog)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 11, 0, 5, 2)
        self.btnDown = QtWidgets.QToolButton(ProjectDialog)
        icon = QtGui.QIcon.fromTheme("go-down")
        self.btnDown.setIcon(icon)
        self.btnDown.setObjectName("btnDown")
        self.gridLayout.addWidget(self.btnDown, 14, 4, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(ProjectDialog)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 17, 3, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)

        self.retranslateUi(ProjectDialog)
        ProjectDialog.setTabOrder(self.lineEdit_Path, self.btnSelectPath)
        ProjectDialog.setTabOrder(self.btnSelectPath, self.spinBoxSampleRate)
        ProjectDialog.setTabOrder(self.spinBoxSampleRate, self.spinBoxFreq)
        ProjectDialog.setTabOrder(self.spinBoxFreq, self.spinBoxBandwidth)
        ProjectDialog.setTabOrder(self.spinBoxBandwidth, self.spinBoxGain)
        ProjectDialog.setTabOrder(self.spinBoxGain, self.txtEdDescription)
        ProjectDialog.setTabOrder(self.txtEdDescription, self.tblParticipants)
        ProjectDialog.setTabOrder(self.tblParticipants, self.btnAddParticipant)
        ProjectDialog.setTabOrder(self.btnAddParticipant, self.btnRemoveParticipant)
        ProjectDialog.setTabOrder(self.btnRemoveParticipant, self.lineEditBroadcastAddress)

    def retranslateUi(self, ProjectDialog):
        _translate = QtCore.QCoreApplication.translate
        ProjectDialog.setWindowTitle(_translate("ProjectDialog", "Create a new project"))
        self.lNewProject.setText(_translate("ProjectDialog", "New Project"))
        self.lblName.setText(_translate("ProjectDialog", "<Name>"))
        self.label_5.setText(_translate("ProjectDialog", "Sps"))
        self.lblNewPath.setText(_translate("ProjectDialog", "<html><head/><body><p><span style=\" font-style:italic;\">Note: A new directory will be created.</span></p></body></html>"))
        self.label_3.setText(_translate("ProjectDialog", "Default frequency:"))
        self.lineEditBroadcastAddress.setToolTip(_translate("ProjectDialog", "<html><head/><body><p>Enter the broadcast address of your protocol in <span style=\" font-weight:600;\">hex</span>. If you do not know what to enter here, just leave the default.</p></body></html>"))
        self.lineEditBroadcastAddress.setText(_translate("ProjectDialog", "ffff"))
        self.label_10.setText(_translate("ProjectDialog", "Default bandwidth:"))
        self.btnAddParticipant.setToolTip(_translate("ProjectDialog", "Add participant"))
        self.btnAddParticipant.setText(_translate("ProjectDialog", "..."))
        self.btnSelectPath.setText(_translate("ProjectDialog", "..."))
        self.label_6.setText(_translate("ProjectDialog", "Hz"))
        self.label_2.setText(_translate("ProjectDialog", "Default sample rate:"))
        self.label.setText(_translate("ProjectDialog", "Choose a path:"))
        self.btnRemoveParticipant.setToolTip(_translate("ProjectDialog", "Remove participant"))
        self.btnRemoveParticipant.setText(_translate("ProjectDialog", "..."))
        self.label_4.setText(_translate("ProjectDialog", "Broadcast address (hex):"))
        self.label_8.setText(_translate("ProjectDialog", "Description:"))
        self.label_12.setText(_translate("ProjectDialog", "Hz"))
        self.label_11.setText(_translate("ProjectDialog", "Default gain:"))
        self.btnUp.setToolTip(_translate("ProjectDialog", "Move selected participants up"))
        self.btnUp.setText(_translate("ProjectDialog", "..."))
        self.lOpenSpectrumAnalyzer.setText(_translate("ProjectDialog", "<html><head/><body><p>Tip: Open <a href=\"open_spectrum_analyzer\">spectrum analyzer</a> to find these values.</p></body></html>"))
        self.label_9.setText(_translate("ProjectDialog", "Participants:"))
        self.btnDown.setToolTip(_translate("ProjectDialog", "Move selected participants down"))
        self.btnDown.setText(_translate("ProjectDialog", "..."))

from urh.ui.KillerDoubleSpinBox import KillerDoubleSpinBox
from urh.ui.views.ParticipantTableView import ParticipantTableView



import array
import ast
import xml.etree.ElementTree as ET
from enum import Enum

from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.util.GenericCRC import GenericCRC
from urh.util.WSPChecksum import WSPChecksum


class ChecksumLabel(ProtocolLabel):
    __slots__ = ("__data_ranges", "checksum", "__category")

    class Category(Enum):
        generic = "generic"
        wsp = "Wireless Short Packet (WSP)"

    def __init__(self, name: str, start: int, end: int, color_index: int, field_type: FieldType,
                 fuzz_created=False, auto_created=False, data_range_start=0):
        assert field_type.function == FieldType.Function.CHECKSUM
        super().__init__(name, start, end, color_index, fuzz_created, auto_created, field_type)

        self.__category = self.Category.generic
        self.__data_ranges = [[data_range_start, self.start]]  # type: list[list[int,int]]
        self.checksum = GenericCRC(polynomial=0)  # type: GenericCRC or WSPChecksum

    def calculate_checksum(self, bits: array.array) -> array.array:
        return self.checksum.calculate(bits)

    def calculate_checksum_for_message(self, message, use_decoded_bits: bool) -> array.array:
        data = array.array("B", [])
        bits = message.decoded_bits if use_decoded_bits else message.plain_bits
        for data_range in self.data_ranges:
            data.extend(bits[data_range[0]:data_range[1]])
        return self.calculate_checksum(data)

    @property
    def data_ranges(self):
        if self.category == self.Category.wsp:
            return [[12, -4]]
        else:
            return self.__data_ranges

    @data_ranges.setter
    def data_ranges(self, value):
        self.__data_ranges = value

    @property
    def is_generic_crc(self):
        return self.category == self.Category.generic

    @property
    def category(self) -> Category:
        return self.__category

    @category.setter
    def category(self, value: Category):
        if value != self.category:
            self.__category = value
            if self.category == self.Category.generic:
                self.checksum = GenericCRC()
            elif self.category == self.Category.wsp:
                self.checksum = WSPChecksum()
            else:
                raise ValueError("Unknown Category")

    def to_label(self, field_type: FieldType) -> ProtocolLabel:
        result = ProtocolLabel(name=self.name, start=self.start, end=self.end - 1, color_index=self.color_index, field_type=field_type,
                               auto_created=self.auto_created, fuzz_created=self.fuzz_created)
        result.apply_decoding = self.apply_decoding
        result.show = self.show
        result.fuzz_me = self.fuzz_me
        result.fuzz_values = self.fuzz_values
        result.display_format_index = self.display_format_index
        return result

    @classmethod
    def from_label(cls, label: ProtocolLabel):
        result = ChecksumLabel(name=label.name, start=label.start, end=label.end - 1, color_index=label.color_index,
                               field_type=FieldType(label.name, FieldType.Function.CHECKSUM),
                               fuzz_created=label.fuzz_created, auto_created=label.auto_created)
        result.apply_decoding = label.apply_decoding
        result.show = label.show
        result.fuzz_me = label.fuzz_me
        result.fuzz_values = label.fuzz_values
        result.display_format_index = label.display_format_index
        return result

    @classmethod
    def from_xml(cls, tag: ET.Element, field_types_by_caption=None):
        lbl = super().from_xml(tag, field_types_by_caption)
        if lbl.field_type is None or lbl.field_type.function != FieldType.Function.CHECKSUM:
            checksum_field_type = next(
                (ft for ft in field_types_by_caption.values() if ft.function == FieldType.Function.CHECKSUM),
                FieldType("checksum", FieldType.Function.CHECKSUM, display_format_index=1))
            lbl.field_type = checksum_field_type
        result = cls.from_label(lbl)
        result.data_ranges = ast.literal_eval(tag.get("data_ranges", "[]"))
        result.category = cls.Category[tag.get("category", "generic")]

        crc_tag = tag.find("crc")
        if crc_tag is not None:
            result.checksum = GenericCRC.from_xml(crc_tag)

        wsp_tag = tag.find("wsp_checksum")
        if wsp_tag is not None:
            result.checksum = WSPChecksum.from_xml(wsp_tag)

        return result

    def to_xml(self):
        result = super().to_xml()
        result.tag = "checksum_label"
        result.attrib.update({"data_ranges": str(self.data_ranges), "category": self.category.name})
        result.append(self.checksum.to_xml())
        return result
